import mock
import unittest
import sys

sys.modules['MySQLdb'] = mock.Mock()

import common  # noqa


class CommonTests(unittest.TestCase):

    @mock.patch.object(common, 'MySQLHelper', spec=common.MySQLHelper)
    def test_get_db_helper(self, mock_helper):
        common.get_db_helper()
        kwargs = {'rpasswdf_template': '/var/lib/mysql/mysql.passwd',
                  'upasswdf_template': '/var/lib/mysql/mysql-{}.passwd',
                  'delete_ondisk_passwd_file': False}
        mock_helper.assert_called_with(**kwargs)



import unittest
from charmtools.generate import (
    graph,
    copy_file,
)
from mock import patch, Mock


class JujuCharmAddTest(unittest.TestCase):
    @patch('charmtools.generate.CharmStore')
    def test_graph(self, store):
        charm = Mock(series='trusty')
        store = store.return_value
        store.provides.return_value = [charm]
        self.assertEqual(graph('nagios', 'requires'), charm)
        store.provides.assert_called_with('nagios')

    @patch('charmtools.generate.CharmStore')
    def test_graph_no_interface(self, store):
        store = store.return_value
        store.provides.return_value = []
        self.assertEqual(graph('nointerface', 'requires'), None)
        store.provides.assert_called_with('nointerface')

    @patch('charmtools.generate.Charm')
    def test_not_charm(self, mcharm):
        mcharm.return_value.is_charm.return_value = False
        self.assertRaises(Exception, copy_file, '1.ex', '/no-charm')



def pre_install():
    """
    Do any setup required before the install hook.
    """
    install_charmhelpers()


def install_charmhelpers():
    """
    Install the charmhelpers library, if not present.
    """
    try:
        import charmhelpers  # noqa
    except ImportError:
        import subprocess
        subprocess.check_call(['apt-get', 'install', '-y', 'python-pip'])
        subprocess.check_call(['pip', 'install', 'charmhelpers'])



#!/usr/bin/python

#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from charmtools.generators import (
    Prompt,
    PromptList,
)


class PromptListTest(TestCase):
    def test_init_empty(self):
        self.assertEqual([], PromptList())

    def test_init_with_data(self):
        data = {
            'symlink': {
                'prompt': 'one file per hook?',
                'default': 'y',
                'type': 'bool',
            }
        }
        pl = PromptList(data)

        self.assertEqual(1, len(pl))
        self.assertIsInstance(pl[0], Prompt)


class PromptTest(TestCase):
    def test_init(self):
        p = Prompt('symlink', 'one file per hook?', 'y')

        self.assertEqual(p.name, 'symlink')
        self.assertEqual(p.prompt, 'one file per hook? ')
        self.assertEqual(p.default, 'y')
        self.assertEqual(p.type_, 'string')

    def test_validate_string(self):
        p1 = Prompt('name', 'your name?', 'Name', 'string')
        p2 = Prompt('name', 'your name?', 'Name', 'str')

        self.assertEqual('Joe', p1.validate('Joe'))
        self.assertEqual('Joe', p2.validate('Joe'))

    def test_validate_int(self):
        p1 = Prompt('age', 'your age?', '0', 'integer')
        p2 = Prompt('age', 'your age?', '0', 'int')

        self.assertEqual(18, p1.validate('18'))
        self.assertEqual(18, p2.validate('18'))

    def test_validate_float(self):
        p1 = Prompt('temp', 'temperature?', '0', 'float')

        self.assertEqual(98.6, p1.validate('98.6'))
        self.assertEqual(100.0, p1.validate('100'))

    def test_validate_bool(self):
        p1 = Prompt('symlink', 'one file per hook?', 'y', 'boolean')
        p2 = Prompt('symlink', 'one file per hook?', 'y', 'bool')

        self.assertTrue(p1.validate('True'))
        self.assertTrue(p1.validate('Yes'))
        self.assertTrue(p1.validate('y'))
        self.assertFalse(p2.validate('False'))
        self.assertFalse(p2.validate('No'))
        self.assertFalse(p2.validate('n'))



#!/usr/bin/env python

import sys
import os

import lib.utils as utils
import lib.ceph_utils as ceph
import lib.cluster_utils as cluster

from charmhelpers.contrib.peerstorage import (
    peer_echo,
)

# CEPH
DATA_SRC_DST = '/var/lib/mysql'
SERVICE_NAME = os.getenv('JUJU_UNIT_NAME').split('/')[0]
POOL_NAME = SERVICE_NAME
LEADER_RES = 'res_mysql_vip'


def cluster_changed():
    # Echo any passwords placed on peer relation
    peer_echo(includes=['.passwd'])


def ha_relation_joined():
    vip = utils.config_get('vip')
    vip_iface = utils.config_get('vip_iface')
    vip_cidr = utils.config_get('vip_cidr')
    corosync_bindiface = utils.config_get('ha-bindiface')
    corosync_mcastport = utils.config_get('ha-mcastport')

    if None in [vip, vip_cidr, vip_iface]:
        utils.juju_log('WARNING',
                       'Insufficient VIP information to configure cluster')
        sys.exit(1)

    # Starting configuring resources.
    init_services = {'res_mysqld': 'mysql'}

    # If the 'ha' relation has been made *before* the 'ceph' relation,
    # it doesn't make sense to make it until after the 'ceph' relation is made
    if not utils.is_relation_made('ceph', 'auth'):
        utils.juju_log('INFO',
                       '*ceph* relation does not exist. '
                       'Not sending *ha* relation data yet')
        return
    else:
        utils.juju_log('INFO',
                       '*ceph* relation exists. Sending *ha* relation data')

        block_storage = 'ceph'

        if utils.config_get('prefer-ipv6'):
            res_mysql_vip = 'ocf:heartbeat:IPv6addr'
            vip_params = 'ipv6addr'
            vip_cidr = '64'
        else:
            res_mysql_vip = 'ocf:heartbeat:IPaddr2'
            vip_params = 'ip'

        resources = {
            'res_mysql_rbd': 'ocf:ceph:rbd',
            'res_mysql_fs': 'ocf:heartbeat:Filesystem',
            'res_mysql_vip': res_mysql_vip,
            'res_mysqld': 'upstart:mysql'}

        rbd_name = utils.config_get('rbd-name')
        resource_params = {
            'res_mysql_rbd': 'params name="%s" pool="%s" user="%s" '
                             'secret="%s"' %
                             (rbd_name, POOL_NAME,
                              SERVICE_NAME, ceph.keyfile_path(SERVICE_NAME)),
            'res_mysql_fs': 'params device="/dev/rbd/%s/%s" directory="%s" '
                            'fstype="ext4" op start start-delay="10s"' %
                            (POOL_NAME, rbd_name, DATA_SRC_DST),
            'res_mysql_vip': 'params "%s"="%s" cidr_netmask="%s" nic="%s"' %
                             (vip_params, vip, vip_cidr, vip_iface),
            'res_mysqld': 'op start start-delay="5s" op monitor interval="5s"'}

        groups = {
            'grp_mysql': 'res_mysql_rbd res_mysql_fs res_mysql_vip res_mysqld'}

        for rel_id in utils.relation_ids('ha'):
            utils.relation_set(rid=rel_id,
                               block_storage=block_storage,
                               corosync_bindiface=corosync_bindiface,
                               corosync_mcastport=corosync_mcastport,
                               resources=resources,
                               resource_params=resource_params,
                               init_services=init_services,
                               groups=groups)


def ha_relation_changed():
    clustered = utils.relation_get('clustered')
    if (clustered and cluster.is_leader(LEADER_RES)):
        utils.juju_log('INFO', 'Cluster configured, notifying other services')
        # Tell all related services to start using the VIP
        for r_id in utils.relation_ids('shared-db'):
            utils.relation_set(rid=r_id,
                               db_host=utils.config_get('vip'))


def ceph_joined():
    utils.juju_log('INFO', 'Start Ceph Relation Joined')
    ceph.install()
    utils.juju_log('INFO', 'Finish Ceph Relation Joined')


def ceph_changed():
    utils.juju_log('INFO', 'Start Ceph Relation Changed')
    auth = utils.relation_get('auth')
    key = utils.relation_get('key')
    use_syslog = utils.relation_get('use_syslog')
    if None in [auth, key]:
        utils.juju_log('INFO', 'Missing key or auth in relation')
        return

    ceph.configure(service=SERVICE_NAME, key=key, auth=auth,
                   use_syslog=use_syslog)

    if cluster.eligible_leader(LEADER_RES):
        sizemb = int(utils.config_get('block-size')) * 1024
        rbd_img = utils.config_get('rbd-name')
        blk_device = '/dev/rbd/%s/%s' % (POOL_NAME, rbd_img)
        rbd_pool_rep_count = utils.config_get('ceph-osd-replication-count')
        ceph.ensure_ceph_storage(service=SERVICE_NAME, pool=POOL_NAME,
                                 rbd_img=rbd_img, sizemb=sizemb,
                                 fstype='ext4', mount_point=DATA_SRC_DST,
                                 blk_device=blk_device,
                                 system_services=['mysql'],
                                 rbd_pool_replicas=rbd_pool_rep_count)
    else:
        utils.juju_log('INFO',
                       'This is not the peer leader. Not configuring RBD.')
        # Stopping MySQL
        if utils.running('mysql'):
            utils.juju_log('INFO', 'Stopping MySQL...')
            utils.stop('mysql')

    # If 'ha' relation has been made before the 'ceph' relation
    # it is important to make sure the ha-relation data is being
    # sent.
    if utils.is_relation_made('ha'):
        utils.juju_log('INFO',
                       '*ha* relation exists. Making sure the ha'
                       ' relation data is sent.')
        ha_relation_joined()
        return

    utils.juju_log('INFO', 'Finish Ceph Relation Changed')


hooks = {
    "ha-relation-joined": ha_relation_joined,
    "ha-relation-changed": ha_relation_changed,
    "ceph-relation-joined": ceph_joined,
    "ceph-relation-changed": ceph_changed,
    "cluster-relation-changed": cluster_changed,
}

utils.do_hooks(hooks)



#!/usr/bin/python

#    Copyright (C) 2013  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import charmtools.bundles
import unittest


class TestCharmProof(unittest.TestCase):
    def setUp(self):
        self.linter = charmtools.bundles.BundleLinter()

    def test_warn_on_charm_urls_without_revisions(self):
        self.linter.validate({
            'services': {
                'memcached': {
                    'charm': 'cs:precise/memcached',
                    'num_units': 1,
                },
            }
        })
        self.assertIn(
            'W: memcached: charm URL should include a revision',
            self.linter.lint)

    def test_no_warning_when_charm_urls_include_revisions(self):
        self.linter.validate({
            'services': {
                'memcached': {
                    'charm': 'cs:precise/memcached-99',
                    'num_units': 1,
                },
            }
        })
        self.assertNotIn(
            'W: memcached: charm URL should include a revision',
            self.linter.lint)

    def test_warn_on_missing_annotations(self):
        self.linter.validate({
            'services': {
                'memcached': {
                    'charm': 'cs:precise/memcached',
                    'num_units': 1,
                },
            }
        })
        self.assertIn(
            'W: memcached: No annotations found, will render '
            'poorly in GUI',
            self.linter.lint)

    def test_no_warning_when_annotations_are_included(self):
        self.linter.validate({
            'my-service': {
                'services': {
                    'memcached': {
                        'charm': 'cs:precise/memcached',
                        'num_units': 1,
                        'annotations': {
                            'gui-x': '821.5',
                            'gui-y': '669.2698359714502',
                        },
                    },
                },
            }})
        self.assertNotIn(
            'W: my-service: memcached: No annotations found, will render '
            'poorly in GUI',
            self.linter.lint)



#!/usr/bin/python

#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
from mock import patch, Mock
from unittest import TestCase

from charmtools.generators import (
    CharmGenerator,
    CharmGeneratorException,
    Prompt,
)

from charmtools.templates.bash import BashCharmTemplate


class CharmGeneratorTest(TestCase):
    def setUp(self):
        class opts(object):
            charmname = 'testcharm'
            charmhome = tempfile.mkdtemp()
            template = 'bash'
            accept_defaults = False

        self.c = CharmGenerator(opts)

    def test_load_plugin(self):
        plugin = self.c._load_plugin()

        self.assertIsInstance(plugin, BashCharmTemplate)

    def test_create_charm_output_path_exists(self):
        os.mkdir(self.c._get_output_path())
        with self.assertRaises(CharmGeneratorException) as e:
            self.c.create_charm()
            self.assertEqual(
                str(e),
                '{} exists. Please move it out of the way.'.format(
                    self.c._get_output_path()))

    def test_create_charm_error(self):
        with patch.object(self.c.plugin, 'create_charm') as create_charm, \
                patch.object(self.c, '_cleanup') as _cleanup:
            create_charm.side_effect = Exception
            with self.assertRaises(Exception):
                self.c.create_charm()
                self.assertTrue(_cleanup.called)

    @patch('charmtools.generators.generator.apt_fill')
    @patch('charmtools.generators.generator.get_maintainer')
    def test_get_metadata(self, get_maintainer, apt_fill):
        get_maintainer.return_value = ('Tester', 'tester@example.com')
        apt_fill.return_value = {
            'summary': 'Charm summary',
            'description': 'Charm description',
        }
        self.assertEqual(
            self.c._get_metadata(), {
                'package': 'testcharm',
                'maintainer': 'Tester <tester@example.com>',
                'summary': 'Charm summary',
                'description': 'Charm description'})

    @patch('__builtin__.raw_input')
    def test_get_user_config_from_prompts(self, raw_input_):
        raw_input_.return_value = 'Yes'
        with patch.object(self.c.plugin, 'prompts') as prompts:
            prompts.return_value = [
                Prompt('symlink', 'symlink hooks?', 'y', 'bool')]
            config = self.c._get_user_config()

            self.assertEqual(config, {'symlink': True})

    def test_get_user_config_from_defaults(self):
        self.c.opts.accept_defaults = True
        with patch.object(self.c.plugin, 'prompts') as prompts:
            prompts.return_value = [
                Prompt('symlink', 'symlink hooks?', 'y', 'bool')]
            config = self.c._get_user_config()

            self.assertEqual(config, {'symlink': True})

    def test_prompt_none(self):
        with patch.object(self.c.plugin, 'configure_prompt') as f:
            f.return_value = None

            self.assertIsNone(self.c._prompt(Mock(), {}))

    def test_prompt_accept_default(self):
        self.c.opts.accept_defaults = True
        prompt = Prompt('symlink', 'symlink hooks?', 'y', 'bool')

        self.assertTrue(self.c._prompt(prompt, {}))

    @patch('__builtin__.raw_input')
    def test_prompt_no_input(self, raw_input_):
        raw_input_.return_value = ''
        prompt = Prompt('symlink', 'symlink hooks?', 'y', 'bool')

        self.assertTrue(self.c._prompt(prompt, {}))

    @patch('__builtin__.raw_input')
    def test_prompt_invalid_input(self, raw_input_):
        raw_input_.side_effect = ['foo', '18']
        prompt = Prompt('age', 'your age?', '42', 'int')

        self.assertEqual(self.c._prompt(prompt, {}), 18)
        self.assertEqual(raw_input_.call_count, 2)

    @patch('__builtin__.raw_input')
    def test_prompt_valid_input(self, raw_input_):
        raw_input_.return_value = 'Joe'
        prompt = Prompt('name', 'your name?', 'Name')

        self.assertEqual(self.c._prompt(prompt, {}), 'Joe')
        self.assertEqual(raw_input_.call_count, 1)

    def test_get_output_path(self):
        path = self.c._get_output_path()
        self.assertEqual(
            path,
            os.path.join(self.c.opts.charmhome, self.c.opts.charmname))

    def test_get_tempdir(self):
        tempdir = self.c._get_tempdir()
        self.assertTrue(os.path.isdir(tempdir))
        self.c._cleanup(tempdir)

    def test_cleanup(self):
        tempdir = self.c._get_tempdir()
        self.c._cleanup(tempdir)
        self.assertFalse(os.path.exists(tempdir))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import blessings
import json
import logging
import os
import requests
import sys
import yaml

import charmtools.build.tactics

from path import path
from collections import OrderedDict
from charmtools import (utils, repofinder, proof)
from charmtools.build import inspector
from charmtools.build.errors import BuildError
from charmtools.build.config import (BuildConfig, DEFAULT_IGNORES)
from charmtools.build.fetchers import (
    InterfaceFetcher,
    LayerFetcher,
    get_fetcher,
    FetchError,
)


log = logging.getLogger("build")


class Configable(object):
    CONFIG_FILE = None

    def __init__(self):
        self._config = BuildConfig()
        self.config_file = None

    @property
    def config(self):
        if self._config.configured:
            return self._config
        if self.config_file and self.config_file.exists():
            self._config.configure(self.config_file)
        return self._config

    @property
    def configured(self):
        return bool(self.config is not None and self.config.configured)


class Fetched(Configable):
    def __init__(self, url, target_repo, name=None):
        super(Fetched, self).__init__()
        self.url = url
        self.target_repo = target_repo / self.NAMESPACE
        self.directory = None
        self._name = name

    @property
    def name(self):
        if self._name:
            return self._name
        if self.url.startswith(self.NAMESPACE):
            return self.url[len(self.NAMESPACE)+1:]
        return self.url

    def __repr__(self):
        return "<{} {}:{}>".format(self.__class__.__name__,
                                   self.url, self.directory)

    def __div__(self, other):
        return self.directory / other

    def fetch(self):
        try:
            fetcher = get_fetcher(self.url)
        except FetchError:
            # We might be passing a local dir path directly
            # which fetchers don't currently  support
            self.directory = path(self.url)
        else:
            if hasattr(fetcher, "path") and fetcher.path.exists():
                self.directory = path(fetcher.path)
            else:
                if not self.target_repo.exists():
                    self.target_repo.makedirs_p()
                self.directory = path(fetcher.fetch(self.target_repo))

        if not self.directory.exists():
            raise BuildError(
                "Unable to locate {}. "
                "Do you need to set {}?".format(
                    self.url, self.ENVIRON))

        self.config_file = self.directory / self.CONFIG_FILE
        if not self.config_file.exists():
            if self.OLD_CONFIG and (self.directory / self.OLD_CONFIG).exists():
                self.config_file = (self.directory / self.OLD_CONFIG)
        self._name = self.config.name
        return self


class Interface(Fetched):
    CONFIG_FILE = "interface.yaml"
    OLD_CONFIG = None
    NAMESPACE = "interface"
    ENVIRON = "INTERFACE_PATH"


class Layer(Fetched):
    CONFIG_FILE = "layer.yaml"
    OLD_CONFIG = "composer.yaml"
    NAMESPACE = "layer"
    ENVIRON = "LAYER_PATH"


class Builder(object):
    """
    Handle the processing of overrides, implements the policy of BuildConfig
    """
    PHASES = ['lint', 'read', 'call', 'sign', 'build']
    HOOK_TEMPLATE_FILE = path('hooks/hook.template')
    DEFAULT_SERIES = 'trusty'

    def __init__(self):
        self.config = BuildConfig()
        self.force = False
        self._name = None
        self._charm = None
        self._top_layer = None
        self.hide_metrics = False

    @property
    def top_layer(self):
        if not self._top_layer:
            self._top_layer = Layer(self.charm, self.deps).fetch()

        return self._top_layer

    @property
    def charm(self):
        return self._charm

    @charm.setter
    def charm(self, value):
        self._charm = path(value)

    @property
    def charm_metadata(self):
        if not hasattr(self, '_charm_metadata'):
            md = path(self.charm) / "metadata.yaml"
            try:
                setattr(
                    self, '_charm_metadata',
                    yaml.load(md.open()) if md.exists() else None)
            except yaml.YAMLError as e:
                log.debug(e)
                raise BuildError("Failed to process {0}. "
                                 "Ensure the YAML is valid".format(md))

        return self._charm_metadata

    @property
    def name(self):
        if self._name:
            return self._name

        # optionally extract name from the top layer
        self._name = str(path(self.charm).abspath().name)

        # however if the current layer has a metadata.yaml we can
        # use its name
        if self.charm_metadata:
            name = self.charm_metadata.get("name")
            if name:
                self._name = name
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def manifest(self):
        return self.target_dir / '.build.manifest'

    def check_series(self):
        """Make sure this is a either a multi-series charm, or we have a
        build series defined. If not, fall back to a default series.

        """
        if self.series:
            return
        if self.charm_metadata and self.charm_metadata.get('series'):
            return
        self.series = self.DEFAULT_SERIES

    def status(self):
        result = {}
        result.update(vars(self))
        for e in ["LAYER_PATH", "INTERFACE_PATH", "JUJU_REPOSITORY"]:
            result[e] = os.environ.get(e)
        return result

    def create_repo(self):
        # Generated output will go into this directory
        base = path(self.output_dir)
        self.repo = (base / (self.series if self.series else 'builds'))
        # And anything it includes from will be placed here
        # outside the series
        self.deps = (base / "deps")
        self.target_dir = (self.repo / self.name)

    def find_or_create_repo(self, allow_create=True):
        # see if output dir is already in a repo, we can use that directly
        if self.output_dir == path(self.charm).abspath():
            # we've indicated in the cmdline that we are doing an inplace
            # update
            if (self.series and
                    self.output_dir.parent.basename() == self.series):
                # we're already in a repo
                self.repo = self.output_dir.parent.parent
                self.deps = (self.repo / "deps")
                self.target_dir = self.output_dir
                return
        if allow_create:
            self.create_repo()
        else:
            raise ValueError("%s doesn't seem valid", self.charm.directory)
        log.info("Destination charm directory: {}".format(self.target_dir))

    @property
    def layers(self):
        layers = []
        for i in self._layers:
            layers.append(i.url)
        for i in self._interfaces:
            layers.append(i.url)
        layers.append("build")
        return layers

    def fetch(self):
        self.target_dir.makedirs_p()
        if not self.top_layer.configured:
            log.warn("The top level layer expects a "
                     "valid layer.yaml file")
        # Manually create a layer object for the output
        self.target = Layer(self.name, self.repo)
        self.target.directory = self.target_dir
        return self.fetch_deps(self.top_layer)

    def fetch_deps(self, layer):
        results = {"layers": [], "interfaces": []}
        self.fetch_dep(layer, results)
        # results should now be a bottom up list
        # of deps. Using the in order results traversal
        # we can build out our plan for each file in the
        # output layer
        results["layers"].append(layer)
        self._layers = results["layers"]
        self._interfaces = results["interfaces"]
        return results

    def fetch_dep(self, layer, results):
        # Recursively fetch and scan layers
        # This returns a plan for each file in the result
        baselayers = layer.config.get('includes', [])
        if not baselayers:
            # no deps, this is possible for any base
            # but questionable for the target
            return

        if isinstance(baselayers, str):
            baselayers = [baselayers]

        for base in baselayers:
            # The order of these commands is important. We only want to
            # fetch something if we haven't already fetched it.
            if base.startswith("interface:"):
                iface = Interface(base, self.deps)
                if iface.name in [i.name for i in results['interfaces']]:
                    continue
                results["interfaces"].append(iface.fetch())
            else:
                base_layer = Layer(base, self.deps)
                if base_layer.name in [i.name for i in results['layers']]:
                    continue
                base_layer.fetch()
                self.fetch_dep(base_layer, results)
                results["layers"].append(base_layer)

    def build_tactics(self, entry, current, config, output_files):
        # Delegate to the config object, it's rules
        # will produce a tactic
        relname = entry.relpath(current.directory)
        current = current.config.tactic(entry, current, self.target, config)
        existing = output_files.get(relname)
        if existing is not None:
            tactic = current.combine(existing)
        else:
            tactic = current
        output_files[relname] = tactic

    def plan_layers(self, layers, output_files):
        config = BuildConfig()
        cfgfn = layers["layers"][0] / BuildConfig.DEFAULT_FILE
        if cfgfn.exists():
            config = config.add_config(
                cfgfn, True)
        else:
            cfgfn = layers["layers"][0] / BuildConfig.OLD_CONFIG
            config = config.add_config(
                cfgfn, True)

        layers["layers"][-1].url = self.name

        for i, layer in enumerate(layers["layers"]):
            log.info("Processing layer: %s", layer.url)
            if i + 1 < len(layers["layers"]):
                next_layer = layers["layers"][i + 1]
                config = config.add_config(
                    next_layer / BuildConfig.DEFAULT_FILE, True)
            list(e for e in utils.walk(layer.directory,
                                       self.build_tactics,
                                       current=layer,
                                       config=config,
                                       output_files=output_files))
        plan = [t for t in output_files.values() if t]
        return plan

    def plan_interfaces(self, layers, output_files, plan):
        # Interface includes don't directly map to output files
        # as they are computed in combination with the metadata.yaml
        if not layers.get('interfaces'):
            return
        metadata_tactic = [tactic for tactic in plan if isinstance(
                           tactic, charmtools.build.tactics.MetadataYAML)]
        if not metadata_tactic:
            raise BuildError('At least one layer must provide metadata.yaml')
        meta = metadata_tactic[0].process()
        if not meta and layers.get('interfaces'):
            raise BuildError(
                'Includes interfaces but no metadata.yaml to bind them')
        elif self.HOOK_TEMPLATE_FILE not in output_files:
            raise BuildError('At least one layer must provide %s',
                             self.HOOK_TEMPLATE_FILE)
        elif not meta:
            log.warn('Empty metadata.yaml')

        template_file = self.target / self.HOOK_TEMPLATE_FILE
        target_config = layers["layers"][-1].config
        specs = []
        used_interfaces = set()
        for role in ("provides", "requires", "peers"):
            for k, v in meta.get(role, {}).items():
                # ex: ["provides", "db", "mysql"]
                specs.append([role, k, v["interface"]])
                used_interfaces.add(v["interface"])

        for iface in layers["interfaces"]:
            if iface.name not in used_interfaces:
                # we shouldn't include something the charm doesn't use
                log.warn("layer.yaml includes {} which isn't "
                         "used in metadata.yaml".format(
                             iface.name))
                continue
            for role, relation_name, interface_name in specs:
                if interface_name != iface.name:
                    continue

                log.info("Processing interface: %s", interface_name)
                # COPY phase
                plan.append(
                    charmtools.build.tactics.InterfaceCopy(
                        iface, relation_name, role,
                        self.target, target_config)
                )
                # Link Phase
                plan.append(
                    charmtools.build.tactics.InterfaceBind(
                        relation_name, iface.url, self.target,
                        target_config, template_file))

    def plan_storage(self, layers, output_files, plan):
        # Storage hooks don't directly map to output files
        # as they are computed in combination with the metadata.yaml
        metadata_tactic = [tactic for tactic in plan if isinstance(
                           tactic, charmtools.build.tactics.MetadataYAML)]
        if not metadata_tactic:
            raise BuildError('At least one layer must provide metadata.yaml')
        meta_tac = metadata_tactic[0]
        meta_tac.process()
        if not meta_tac.storage:
            return
        if self.HOOK_TEMPLATE_FILE not in output_files:
            raise BuildError('At least one layer must provide %s',
                             self.HOOK_TEMPLATE_FILE)

        template_file = self.target / self.HOOK_TEMPLATE_FILE
        target_config = layers["layers"][-1].config
        for name, owner in meta_tac.storage.items():
            plan.append(
                charmtools.build.tactics.StorageBind(
                    name, owner, self.target,
                    target_config, template_file))

    def formulate_plan(self, layers):
        """Build out a plan for each file in the various
        layers, taking into account config at each layer"""
        output_files = OrderedDict()
        self.plan = self.plan_layers(layers, output_files)
        self.plan_interfaces(layers, output_files, self.plan)
        self.plan_storage(layers, output_files, self.plan)
        if self.hide_metrics is not True:
            self.post_metrics(layers)
        return self.plan

    def post_metrics(self, layers):
        url = "/".join((self.interface_service,
                        "api/v1/metrics/"))
        data = {"kind": "build",
                "layers": [l.url for l in layers["layers"]],
                "interfaces": [i.url for i in layers["interfaces"]]}
        try:
            requests.post(url, json.dumps(data).encode('utf-8'), timeout=10)
        except requests.exceptions.RequestException:
            log.warning("Unable to post usage metrics")

    def exec_plan(self, plan=None, layers=None):
        signatures = {}
        cont = True
        for phase in self.PHASES:
            for tactic in plan:
                if phase == "lint":
                    cont &= tactic.lint()
                    if cont is False and self.force is not True:
                        return
                elif phase == "read":
                    # We use a read (into memory phase to make layer comps
                    # simpler)
                    tactic.read()
                elif phase == "call":
                    tactic()
                elif phase == "sign":
                    sig = tactic.sign()
                    if sig:
                        signatures.update(sig)
        new_repo = not self.manifest.exists()
        if new_repo:
            added, changed, removed = set(), set(), set()
        else:
            ignores = utils.ignore_matcher(DEFAULT_IGNORES)
            added, changed, _ = utils.delta_signatures(self.manifest, ignores)
            removed = self.clean_removed(signatures)
        # write out the sigs
        if "sign" in self.PHASES:
            self.write_signatures(signatures, layers)
        if self.report:
            self.write_report(new_repo, added, changed, removed)

    def write_signatures(self, signatures, layers):
        signatures['.build.manifest'] = ["build", 'dynamic', 'unchecked']
        self.manifest.write_text(json.dumps(dict(
            signatures=signatures,
            layers=layers,
        ), indent=2, sort_keys=True))

    def generate(self):
        layers = self.fetch()
        self.formulate_plan(layers)
        self.exec_plan(self.plan, self.layers)

    def validate(self):
        self._validate_charm_repo()

        if not self.manifest.exists():
            return [], [], []
        ignorer = utils.ignore_matcher(DEFAULT_IGNORES)
        a, c, d = utils.delta_signatures(self.manifest, ignorer)

        for f in a:
            log.warn(
                "Added unexpected file, should be in a base layer: %s", f)
        for f in c:
            log.warn(
                "Changed file owned by another layer: %s", f)
        for f in d:
            log.warn(
                "Deleted a file owned by another layer: %s", f)
        if a or c or d:
            if self.force is True:
                log.info(
                    "Continuing with known changes to target layer. "
                    "Changes will be overwritten")
            else:
                raise BuildError(
                    "Unable to continue due to unexpected modifications "
                    "(try --force)")

        return a, c, d

    def _validate_charm_repo(self):
        if 'repo' not in self.top_layer.config:
            msg = 'Please add a `repo` key to your {}'.format(
                self.top_layer.config_file.name)
            recommended_repo = repofinder.get_recommended_repo(self.charm)
            if recommended_repo:
                msg += ', e.g. repo: {}'.format(recommended_repo)
            else:
                msg += ', with a url from which your layer can be cloned.'
            log.warn(msg)

    def __call__(self):
        self.find_or_create_repo()

        log.debug(json.dumps(
            self.status(), indent=2, sort_keys=True, default=str))
        self.validate()
        self.generate()

    def inspect(self):
        self.charm = path(self.charm).abspath()
        inspector.inspect(self.charm, force_styling=self.force_raw)

    def normalize_outputdir(self):
        od = path(self.charm).abspath()
        repo = os.environ.get('JUJU_REPOSITORY')
        if repo:
            repo = path(repo)
            if repo.exists():
                od = repo
        elif ":" in od:
            od = od.basename
        log.info("Composing into {}".format(od))
        self.output_dir = od

    def clean_removed(self, signatures):
        """
        Clean up any files that were accounted for in the previous build
        manifest but which have been removed in the current set of sigs.
        """
        old_sigs = json.loads(self.manifest.text())['signatures']
        old_files = set(old_sigs.keys()) - {'.build.manifest'}
        new_files = set(signatures.keys())
        removed = old_files - new_files
        for filename in removed:
            filepath = self.target_dir / filename
            filepath.remove_p()
        return removed

    def write_report(self, new_repo, added, changed, removed):
        """
        Log messages indicating what changed with this (re)build.
        """
        log.info('')
        log.info('---------------------------------------')
        log.info('              Build Report')
        log.info('---------------------------------------')

        if new_repo:
            log.info('New build; all files were modified.')
            return
        elif any([added, changed, removed]):
            sigils = ['+', ' ', '-']
            for sigil, filenames in zip(sigils, [added, changed, removed]):
                for filename in filenames:
                    log.info(' {} {}'.format(sigil, filename))
        else:
            log.info('No new changes; no files were modified.')


def configLogging(build):
    global log
    logging.captureWarnings(True)
    clifmt = utils.ColoredFormatter(
        blessings.Terminal(),
        '%(name)s: %(message)s')
    root_logger = logging.getLogger()
    clihandler = logging.StreamHandler(sys.stdout)
    clihandler.setFormatter(clifmt)
    if isinstance(build.log_level, str):
        build.log_level = build.log_level.upper()
    root_logger.setLevel(build.log_level)
    log.setLevel(build.log_level)
    root_logger.addHandler(clihandler)
    requests_logger = logging.getLogger("requests")
    requests_logger.setLevel(logging.WARN)
    urllib_logger = logging.getLogger("urllib3")
    urllib_logger.setLevel(logging.CRITICAL)


def inspect(args=None):
    build = Builder()
    parser = argparse.ArgumentParser(
        description='inspect the layers of a built charm')
    parser.add_argument('-r', '--force-raw', action="store_true",
                        help="Force raw output (color)")
    parser.add_argument('-l', '--log-level', default=logging.INFO)
    parser.add_argument('charm', nargs="?", default=".", type=path)
    utils.add_plugin_description(parser)
    # Namespace will set the options as attrs of build
    parser.parse_args(args, namespace=build)
    configLogging(build)
    build.inspect()


def deprecated_main():
    namemap = {
        'compose': 'build',
        'generate': 'build',
        'refresh': 'build',
        'inspect': 'layers',
    }
    cmd = sys.argv[0]
    if "-" in cmd:
        old = cmd.rsplit('-', 1)[-1]
    else:
        old = sys.argv[1]
    new = namemap[old]

    class MockBuild(object):
        log_level = 'INFO'

    configLogging(MockBuild)

    msg = "{} has been deprecated, please use {}".format(old, new)
    if '--description' in sys.argv:
        print(msg)
    else:
        log.critical(msg)


def main(args=None):
    build = Builder()
    parser = argparse.ArgumentParser(
        description="build a charm from layers and interfaces",
        formatter_class=argparse.RawDescriptionHelpFormatter,)
    parser.add_argument('-l', '--log-level', default=logging.INFO)
    parser.add_argument('-f', '--force', action="store_true")
    parser.add_argument('-o', '--output-dir', type=path)
    parser.add_argument('-s', '--series', default=None)
    parser.add_argument('--hide-metrics', dest="hide_metrics",
                        default=False, action="store_true")
    parser.add_argument('--interface-service',
                        default="http://interfaces.juju.solutions")
    parser.add_argument('--no-local-layers', action="store_true",
                        help="Don't use local layers when building. "
                        "Forces included layers to be downloaded "
                        "from the interface service.")
    parser.add_argument('-n', '--name',
                        help="Build a charm of 'name' from 'charm'")
    parser.add_argument('-r', '--report', action="store_true",
                        help="Show post-build report of changes")
    parser.add_argument('charm', nargs="?", default=".", type=path)
    utils.add_plugin_description(parser)
    # Namespace will set the options as attrs of build
    parser.parse_args(args, namespace=build)
    if build.charm == "help":
        parser.print_help()
        raise SystemExit(0)

    # Monkey patch in the domain for the interface webservice
    InterfaceFetcher.INTERFACE_DOMAIN = build.interface_service
    LayerFetcher.INTERFACE_DOMAIN = build.interface_service

    InterfaceFetcher.NO_LOCAL_LAYERS = build.no_local_layers

    configLogging(build)

    try:
        if not build.output_dir:
            build.normalize_outputdir()
        if not build.series:
            build.check_series()

        build()

        lint, exit_code = proof.proof(build.target_dir, False, False)
        llog = logging.getLogger("proof")

        if not lint:
            llog.info('OK!')

        for line in lint:
            if line[0] == "I":
                llog.info(line)
            elif line[0] == "W":
                llog.warn(line)
            elif line[0] == "E":
                llog.error(line)
    except (BuildError, FetchError) as e:
        log.error(*e.args)
        raise SystemExit(1)


if __name__ == '__main__':
    main()



#
# Copyright 2014 Canonical Ltd.
#
# Authors:
#  Edward Hope-Morley <opentastic@gmail.com>
#

import time

from charmhelpers.core.hookenv import (
    log,
    INFO,
)


def retry_on_exception(num_retries, base_delay=0, exc_type=Exception):
    """If the decorated function raises exception exc_type, allow num_retries
    retry attempts before raise the exception.
    """
    def _retry_on_exception_inner_1(f):
        def _retry_on_exception_inner_2(*args, **kwargs):
            retries = num_retries
            multiplier = 1
            while True:
                try:
                    return f(*args, **kwargs)
                except exc_type:
                    if not retries:
                        raise

                delay = base_delay * multiplier
                multiplier += 1
                log("Retrying '%s' %d more times (delay=%s)" %
                    (f.__name__, retries, delay), level=INFO)
                retries -= 1
                if delay:
                    time.sleep(delay)

        return _retry_on_exception_inner_2

    return _retry_on_exception_inner_1



#!/usr/bin/python

#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from fnmatch import fnmatch
import inspect
import os
import yaml

from .prompt import PromptList


class CharmTemplate(object):
    """Base plugin for creating a new charm."""

    skip_parsing = ['README.ex', '*.pyc']

    def skip_template(self, filename):
        """Return True if file should not be processed for template
        variable substitution.

        """
        return any(fnmatch(filename, pat) for pat in self.skip_parsing)

    def prompts(self):
        """Return a list :class:`Prompt` objects that will be used for
        configuring the charm created by this plugin.

        """
        return PromptList(self.config().get('prompts'))

    def config(self):
        """Return default configuration for this template, loaded from a
        config.yaml file.

        This is a sample config.yaml that configures one user prompt::

            prompts:
              symlink:
                prompt: Symlink all hooks to one python source file? [yN]
                default: n
                type: boolean

        """
        path = self.config_path()
        if os.path.exists(path):
            with open(path, 'r') as f:
                return yaml.safe_load(f.read())
        return {}

    def config_path(self):
        """Return path to the config yaml file for this template.

        """
        class_file = inspect.getfile(self.__class__)
        return os.path.join(os.path.dirname(class_file), 'config.yaml')

    def create_charm(self, config, output_path):
        """Create charm files

        :param config: dict of config values gathered interactively from the
            user, loaded from a config file, or as a result of accepting all
            configuration defaults.

        :param output_path: directory into which all charm files and dirs
            should be written.

        """
        raise NotImplementedError

    def configure_prompt(self, prompt, config):
        """Reconfigure a prompt based on already-gathered config options

        Called right before ``prompt`` is rendered to the user or before
        the default value for the prompt is accepted. This gives the plugin
        a chance to reconfigure a prompt in any way necessary based on the
        results of prior prompts (contained in ``config``), including
        changing its :attr:`prompt` or :attr:`default`.

        Valid return values are the original prompt (modified or not), an
        entirely new :class:`Prompt` object, or None if this prompt should
        be skipped altogether.

        :param config: dict of all configuration values that have been set
            prior to this prompt being called.

        """
        return prompt

    def validate_input(self, input_value, prompt, config):
        """Return the (possibly modified) validated input value, or raise
        ValueError with a message explaining why the value is invalid.

        :param input_value: str entered by user
        :param prompt: :class:`Prompt` object that elicited this input
        :param config: dict of all configuration values that have been set
            prior to this prompt being called.

        """
        return prompt.validate(input_value)



#!/usr/bin/python
#
# Create relations between a shared database to many peers.
# Join does nothing.   Peer requests access to $DATABASE from $REMOTE_HOST.
# It's up to the hooks to ensure database exists, peer has access and
# clean up grants after a broken/departed peer (TODO)
#
# Author: Adam Gandelman <adam.gandelman@canonical.com>

import subprocess
import json
import lib.utils as utils
import lib.cluster_utils as cluster

from common import get_db_helper
from charmhelpers.core import hookenv
from charmhelpers.contrib.network.ip import (
    get_ipv6_addr
)

LEADER_RES = 'res_mysql_vip'


def pwgen():
    return str(subprocess.check_output(['pwgen', '-s', '16'])).strip()


def relation_get():
    return json.loads(subprocess.check_output(['relation-get',
                                               '--format',
                                               'json']))


def unit_sorted(units):
    """Return a sorted list of unit names."""
    return sorted(
        units, lambda a, b: cmp(int(a.split('/')[-1]), int(b.split('/')[-1])))


def get_unit_addr(relid, unitid):
    return hookenv.relation_get(attribute='private-address',
                                unit=unitid,
                                rid=relid)


def shared_db_changed():
    if not cluster.eligible_leader(LEADER_RES):
        utils.juju_log('INFO',
                       'MySQL service is peered, bailing shared-db relation'
                       ' as this service unit is not the leader')
        return

    if utils.config_get('prefer-ipv6'):
        local_hostname = get_ipv6_addr(exc_list=[utils.config_get('vip')])[0]
    else:
        local_hostname = utils.unit_get('private-address')

    settings = relation_get()
    singleset = set([
        'database',
        'username',
        'hostname'])

    db_helper = get_db_helper()

    if singleset.issubset(settings):
        # Process a single database configuration
        hostname = settings['hostname']
        database = settings['database']
        username = settings['username']

        # Hostname can be json-encoded list of hostnames
        try:
            hostname = json.loads(hostname)
        except ValueError:
            hostname = [hostname]

        for host in hostname:
            password = db_helper.configure_db(host, database, username)

        allowed_units = db_helper.get_allowed_units(database, username)
        allowed_units = unit_sorted(allowed_units)
        allowed_units = ' '.join(allowed_units)

        if cluster.is_clustered():
            db_host = utils.config_get("vip")
        else:
            db_host = local_hostname

        utils.relation_set(db_host=db_host,
                           password=password,
                           allowed_units=allowed_units)
    else:
        # Process multiple database setup requests.
        # from incoming relation data:
        #  nova_database=xxx nova_username=xxx nova_hostname=xxx
        #  quantum_database=xxx quantum_username=xxx quantum_hostname=xxx
        # create
        # {
        #   "nova": {
        #        "username": xxx,
        #        "database": xxx,
        #        "hostname": xxx
        #    },
        #    "quantum": {
        #        "username": xxx,
        #        "database": xxx,
        #        "hostname": xxx
        #    }
        # }
        #
        databases = {}
        for k, v in settings.iteritems():
            db = k.split('_')[0]
            x = '_'.join(k.split('_')[1:])
            if db not in databases:
                databases[db] = {}
            databases[db][x] = v

        return_data = {}
        for db in databases:
            if singleset.issubset(databases[db]):
                database = databases[db]['database']
                hostname = databases[db]['hostname']
                username = databases[db]['username']

                try:
                    # Can be json-encoded list of hostnames
                    hostname = json.loads(hostname)
                except ValueError:
                    # Otherwise expected to be single hostname
                    hostname = [hostname]

                for host in hostname:
                    password = db_helper.configure_db(host, database, username)

                a_units = db_helper.get_allowed_units(database, username)
                a_units = ' '.join(unit_sorted(a_units))
                return_data['%s_allowed_units' % (db)] = a_units

                return_data['%s_password' % (db)] = password

        if len(return_data) > 0:
            utils.relation_set(**return_data)
        if not cluster.is_clustered():
            utils.relation_set(db_host=local_hostname)
        else:
            utils.relation_set(db_host=utils.config_get("vip"))


hooks = {"shared-db-relation-changed": shared_db_changed}

utils.do_hooks(hooks)



#!/usr/bin/python

#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from mock import Mock
from unittest import TestCase

from charmtools.generators import (
    CharmTemplate,
)


class CharmTemplateTest(TestCase):
    def test_create_charm(self):
        t = CharmTemplate()

        self.assertRaises(NotImplementedError, t.create_charm, {}, '.')

    def test_configure_prompt(self):
        t = CharmTemplate()
        prompt = Mock()

        self.assertEqual(prompt, t.configure_prompt(prompt, {}))

    def test_validate_input(self):
        t = CharmTemplate()
        prompt = Mock()
        t.validate_input('value', prompt, {})

        prompt.validate.assert_called_once_with('value')



import logging
import unittest

from charmtools.build.config import BuildConfig


class TestConfig(unittest.TestCase):
    def test_rget(self):
        c = BuildConfig()
        c['a'] = 1
        c = c.new_child()
        c['a'] = 99
        c['b'] = "alpha"
        self.assertEqual(c.get('a'), 99)
        self.assertEqual(c.get('b'), "alpha")
        self.assertEqual(c.rget('a'), [99, 1])

    def test_tactics(self):
        # configure from empty and a layer with tactics
        c = BuildConfig()
        c._tactics = ['a', 'b', 'c']
        c = c.new_child()
        c._tactics = ['d', 'c']
        self.assertEqual(c.tactics()[:5], ['d', 'c', 'a', 'b', 'c'])


if __name__ == '__main__':
    logging.basicConfig()
    unittest.main()



#!/usr/bin/python

#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


def get_validator(type_):
    return {
        'str': str,
        'string': str,
        'int': int,
        'integer': int,
        'float': float,
        'bool': boolean_validator,
        'boolean': boolean_validator,
    }[type_]


def boolean_validator(s):
    return s and s.lower() == 'true' or s.lower()[0] == 'y'


class PromptList(list):
    def __init__(self, prompt_dicts=None):
        prompts = []
        for k, v in (prompt_dicts or {}).items():
            prompts.append(Prompt(
                k,
                v['prompt'],
                v['default'],
                v.get('type', 'string'),
            ))
        super(PromptList, self).__init__(prompts)


class Prompt(object):
    def __init__(self, name, prompt, default, type_='string'):
        self.name = name
        self.prompt = prompt.strip() + ' '
        self.default = default
        self.type_ = type_

    def validate(self, value):
        """Return the (possibly modified) validated value, or raise an
        Exception with a message explaining why the value is invalid.

        """
        return get_validator(self.type_)(value)



#!/usr/bin/env python

import os
import sys
import uuid
import MySQLdb
from charmhelpers.core.hookenv import (
    log,
    relations_of_type,
    Hooks, UnregisteredHookError
)
from charmhelpers.contrib.charmsupport.nrpe import NRPE
from common import get_db_cursor


hooks = Hooks()


def nagios_password():
    PASSFILE = os.path.join('/var', 'lib', 'mysql', 'nagios.passwd')
    if not os.path.isfile(PASSFILE):
        password = str(uuid.uuid4())
        with open(PASSFILE, 'w') as f:
            f.write(password)
        os.chmod(PASSFILE, 0600)
    else:
        with open(PASSFILE, 'r') as rpw:
            password = rpw.read()
    return password


def add_nagios_user():
    c = get_db_cursor()
    if c.execute("SELECT User from mysql.user where User='nagios'"):
        log('User nagios already exists, skipping')
    else:
        log('Creating "nagios" database user')
        password = nagios_password()
        c.execute("CREATE USER 'nagios'@'localhost' IDENTIFIED BY '{}'"
                  "".format(password))

    try:
        c.execute("SHOW GRANTS FOR 'nagios'@'localhost'")
        grants = [i[0] for i in c.fetchall()]
    except MySQLdb.OperationalError:
        grants = []

    for grant in grants:
        if "GRANT PROCESS ON *.*" in grant:
            log('User already has permissions, skipping')
            c.close()
            return
    log('Granting "PROCESS" privilege to nagios user')
    c.execute("GRANT PROCESS ON *.* TO 'nagios'@'localhost'")
    c.close()


def update_nrpe_checks():
    log('Refreshing nrpe checks')
    # Find out if nrpe set nagios_hostname
    hostname = None
    for rel in relations_of_type('nrpe-external-master'):
        if 'nagios_hostname' in rel:
            hostname = rel['nagios_hostname']
            break
    nrpe = NRPE(hostname=hostname)
    nrpe.add_check(
        shortname='mysql_proc',
        description='Check MySQL process',
        check_cmd='check_procs -c 1:1 -C mysqld'
    )
    nrpe.add_check(
        shortname='mysql',
        description='Check MySQL connectivity',
        check_cmd='check_mysql -u nagios -p {}'.format(nagios_password())
    )
    nrpe.write()


@hooks.hook('nrpe-external-master-relation-changed')
@hooks.hook('nrpe-external-master-relation-joined')
def nrpe_relation():
    add_nagios_user()
    update_nrpe_checks()


if __name__ == '__main__':
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as e:
        log('Unknown hook {} - skipping.'.format(e))



import glob
import os
import re
import yaml

from linter import Linter
import jujubundlelib.validation


charm_url_includes_id = re.compile(r'-\d+$').search


class BundleLinter(Linter):
    def validate(self, data):
        """Supplement jujubundlelib validation with some extra checks.

        """
        if 'series' not in data and 'inherits' not in data:
            self.info("No series defined")

        if 'services' in data:
            for svc, sdata in data['services'].items():
                if 'annotations' not in sdata:
                    self.warn('%s: No annotations found, will render '
                              'poorly in GUI' % svc)
                if ('charm' in sdata and
                        not charm_url_includes_id(sdata['charm'] or '')):
                    self.warn(
                        '%s: charm URL should include a revision' % svc)
        else:
            if 'inherits' not in data:
                self.err("No services defined")

    def proof(self, bundle):
        data = bundle.bundle_file()
        if not bundle.is_v4(data):
            self.err(
                'This bundle format is no longer supported. See '
                'https://jujucharms.com/docs/stable/charms-bundles '
                'for the supported format.')
            return

        readmes = glob.glob(os.path.join(bundle.bundle_path, 'README*'))
        if len(readmes) < 1:
            self.warn('No readme file found')

        errors = jujubundlelib.validation.validate(data)
        for error in errors:
            self.err(error)
        self.validate(data)


class Bundle(object):
    def __init__(self, path, debug=False):
        self.bundle_path = os.path.abspath(path)
        self.supported_files = [
            'bundle.yaml'
        ]
        self.debug = debug
        if not self.is_bundle():
            raise Exception('Not a bundle')

    def is_bundle(self):
        for f in self.supported_files:
            if os.path.isfile(os.path.join(self.bundle_path, f)):
                break
        else:
            return False

        if os.path.isfile(os.path.join(self.bundle_path, 'metadata.yaml')):
            return False

        return True

    def is_v4(self, data=None):
        if data is None:
            data = self.bundle_file()
        v4_keys = {'services', 'relations', 'machines', 'series'}
        bundle_keys = set(data.keys())
        return bool(v4_keys & bundle_keys)

    def bundle_file(self, parse=True):
        for f in self.supported_files:
            if os.path.isfile(os.path.join(self.bundle_path, f)):
                with open(os.path.join(self.bundle_path, f)) as d:
                    return yaml.safe_load(d.read()) if parse else d.read()

        raise Exception('No bundle.json or bundle.yaml file found')

    def proof(self):
        lint = BundleLinter(self.debug)
        lint.proof(self)
        return lint.lint, lint.exit_code

    def promulgate(self):
        pass



from unittest import TestCase
from charmtools import utils
from StringIO import StringIO


class TestUtils(TestCase):

    def test_delta_python(self):
        a = StringIO("""
        def foo(n):
            return n * 2


        @when('db.ready')
        def react(db):
            print db
        """)

        b = StringIO("""
        def foo(n):
            return n * 2


        @when('db.ready', 'bar')
        def react(db):
            print db
        """)

        result = StringIO()
        t = utils.TermWriter(fp=result)
        rc = utils.delta_python_dump(a, b, utils.REACTIVE_PATTERNS,
                                     context=3,
                                     term=t,
                                     from_name="Alpha",
                                     to_name="Beta")
        # return code here indicates that there was a diff
        self.assertFalse(rc)
        result.seek(0)
        output = result.read()
        self.assertIn("Alpha", output)
        self.assertIn("Beta", output)
        self.assertIn("@when('db.ready'", output)
        self.assertIn("bar", output)



#!/usr/bin/python
#
#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import os.path as path
import time
import shutil
import subprocess
import tempfile

from Cheetah.Template import Template
from stat import ST_MODE

from charmtools.generators import (
    CharmTemplate,
)

log = logging.getLogger(__name__)


class ReactivePythonCharmTemplate(CharmTemplate):
    """Creates a reactive, layered python-based charm"""

    # _EXTRA_FILES is the list of names of files present in the git repo
    # we don't want transferred over to the charm template:
    _EXTRA_FILES = ["README.md", ".git", ".gitmodules"]

    _TEMPLATE_URL = "https://github.com/juju-solutions/template-reactive-python"

    def create_charm(self, config, output_dir):
        self._clone_template(config, output_dir)

        for root, dirs, files in os.walk(output_dir):
            for outfile in files:
                if self.skip_template(outfile):
                    continue

                self._template_file(config, path.join(root, outfile))

    def _template_file(self, config, outfile):
        if path.islink(outfile):
            return

        mode = os.stat(outfile)[ST_MODE]
        t = Template(file=outfile, searchList=(config))
        o = tempfile.NamedTemporaryFile(
            dir=path.dirname(outfile), delete=False)
        os.chmod(o.name, mode)
        o.write(str(t))
        o.close()
        backupname = outfile + str(time.time())
        os.rename(outfile, backupname)
        os.rename(o.name, outfile)
        os.unlink(backupname)

    def _clone_template(self, config, output_dir):
        cmd = "git clone {} {}".format(
            self._TEMPLATE_URL, output_dir
        )

        try:
            subprocess.check_call(cmd.split())
        except OSError as e:
            raise Exception(
                "The below error has occurred whilst attempting to clone"
                "the charm template. Please make sure you have git"
                "installed on your system.\n" + e
            )

        # iterate and remove all the unwanted files from the git repo:
        for item in [path.join(output_dir, i) for i in self._EXTRA_FILES]:
            if not path.exists(item):
                continue

            if path.isdir(item) and not path.islink(item):
                shutil.rmtree(item)
            else:
                os.remove(item)

        # rename handlers.py to <charm-name>.py
        new_name = '%s.py' % config['metadata']['package'].replace('-', '_')
        os.rename(os.path.join(output_dir, 'reactive', 'handlers.py'),
                  os.path.join(output_dir, 'reactive', new_name))



#!/usr/bin/python

# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import argparse

from Cheetah.Template import Template

from cli import parser_defaults
from charms import Charm
from charmstore import CharmStore
from charmstore.error import CharmNotFound
from . import utils

TPL_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'templates')
CHARM_TPL = os.path.join(TPL_DIR, 'charm')


def graph(interface, endpoint, series='trusty'):
    matches = {
        'requires': 'provides',
        'provides': 'requires',
    }
    match = matches[endpoint]
    c = CharmStore()
    try:
        charms = getattr(c, match)(interface)
    except CharmNotFound:
        return None
    charms = [c for c in charms if c.series == series]
    if charms:
        return charms[0]
    else:
        return None


def copy_file(tpl_file, charm_dir, is_bundle=False, debug=False):
    c = Charm(charm_dir)

    if not c.is_charm():
        raise Exception('%s is not a charm' % charm_dir)

    shutil.copy(os.path.join(CHARM_TPL, tpl_file), charm_dir)


def tests(charm_dir, is_bundle=False, debug=False, series='trusty'):
    c = Charm(charm_dir)

    if not c.is_charm():
        raise Exception('Not a Charm')

    mdata = c.metadata()

    interfaces = {}
    deploy = [mdata['name']]
    relations = []

    for rel_type in ['provides', 'requires']:
        if rel_type in mdata:
            interfaces[rel_type] = {}
            for rel, data in mdata[rel_type].iteritems():
                iface = data['interface']
                if iface and iface not in interfaces[rel_type]:
                    r = graph(iface, rel_type, series=series)
                    # If we dont find an interface, do nothing
                    if r is None:
                        continue
                    interfaces[rel_type][iface] = r
                    deploy.append(r.name)

                relations.append(['%s:%s' % (mdata['name'], rel), r.name])

    t = Template(file=os.path.join(TPL_DIR, 'tests', '99-autogen.tpl'),
                 searchList=[{'deploy': deploy, 'relate': relations,
                              'series': series}])

    if not os.path.exists(os.path.join(charm_dir, 'tests')):
        os.mkdir(os.path.join(charm_dir, 'tests'))

    with open(os.path.join(charm_dir, 'tests', '99-autogen'), 'w') as f:
        f.write(str(t))

    if not os.path.exists(os.path.join(charm_dir, 'tests', '00-setup')):
        with open(os.path.join(charm_dir, 'tests', '00-setup'), 'w') as f:
            f.write("""#!/bin/bash

sudo add-apt-repository ppa:juju/stable -y
sudo apt-get update
sudo apt-get install amulet python3-requests -y
""")

    os.chmod(os.path.join(charm_dir, 'tests', '99-autogen'), 0755)
    os.chmod(os.path.join(charm_dir, 'tests', '00-setup'), 0755)


def parser(args=None):
    parser = argparse.ArgumentParser(
        description='add icon, readme, or tests to a charm')
    parser.add_argument('subcommand', choices=['tests', 'readme', 'icon'],
                        help='Which type of generator to run')
    utils.add_plugin_description(parser)
    parser = parser_defaults(parser)
    return parser.parse_known_args(args)


def tests_parser(args):
    # This bites, need an argparser experter
    parser = argparse.ArgumentParser(description="add tests to a charm")
    parser.add_argument('--series', '-s', default='trusty',
                        help='Series for the generated test')
    return parser.parse_args(args)


def main():
    a, extra = parser()
    if a.subcommand == 'tests':
        opts = tests_parser(extra)
        tests(os.getcwd(), is_bundle=a.bundle, debug=a.debug,
              series=opts.series)
    elif a.subcommand == 'readme':
        copy_file('README.ex', os.getcwd(), is_bundle=a.bundle, debug=a.debug)
    elif a.subcommand == 'icon':
        copy_file('icon.svg', os.getcwd(), is_bundle=a.bundle, debug=a.debug)
    else:
        raise Exception('No subcommand found')


if __name__ == '__main__':
    main()



from .tactics import DEFAULT_TACTICS, load_tactic


import pathspec
from ruamel import yaml
import logging
from path import path
from otherstuf import chainstuf

DEFAULT_IGNORES = [
    ".bzr/",
    ".git/",
    "**/.ropeproject/",
    "*.pyc",
    "*~",
    ".tox/",
    "build/",
]


class BuildConfig(chainstuf):
    """Defaults for controlling the generator, each layer in
    the inclusion graph can provide values, including things
    like overrides, or warnings if things are overridden that
    shouldn't be.
    """
    DEFAULT_FILE = "layer.yaml"
    OLD_CONFIG = "composer.yaml"

    def __init__(self, *args, **kwargs):
        super(BuildConfig, self).__init__(*args, **kwargs)
        self['_tactics'] = []
        self.configured = False

    def __getattr__(self, key):
        return self[key]

    def rget(self, key):
        """Combine all the results from all the layers into a single iter"""
        result = []
        for m in self.maps:
            r = m.get(key)
            if r:
                if isinstance(r, (list, tuple)):
                    result.extend(r)
                else:
                    result.append(r)
        return result

    def configure(self, config_file, allow_missing=False):
        config_file = path(config_file)
        data = None
        if not config_file.exists() and not allow_missing:
            raise OSError("Missing Config File {}".format(config_file))
        try:
            if config_file.exists() and config_file.text().strip() != "":
                data = yaml.load(config_file.open())
                self.configured = True
        except yaml.parser.ParserError:
            logging.critical("Malformed Config file: {}".format(config_file))
            raise
        if data:
            self.update(data)
        # look at any possible imports and use them to build tactics
        tactics = self.get('tactics')
        basedir = config_file.dirname()
        if tactics:
            for name in tactics:
                tactic = load_tactic(name, basedir)
                self._tactics.append(tactic)
        return self

    @classmethod
    def from_config(cls, config_file, allow_missing=False):
        c = cls()
        c.configure(config_file, allow_missing)
        return c

    def add_config(self, config_file, allow_missing=False):
        c = self.new_child()
        c.configure(config_file, allow_missing)
        return c

    @property
    def name(self):
        return self.get('name')

    @property
    def ignores(self):
        return self.rget('ignore') + DEFAULT_IGNORES

    def tactics(self):
        # XXX: combine from config layer
        return self.rget('_tactics') + DEFAULT_TACTICS

    def tactic(self, entity, current, target, next_config):
        # Produce a tactic for the entity in question
        # These will be accumulate through the layers
        # and executed later
        bd = current.directory
        # Ignore handling
        if next_config:
            spec = pathspec.PathSpec.from_lines(pathspec.GitIgnorePattern,
                                                next_config.ignores)
            p = entity.relpath(bd)
            matches = spec.match_files((p,))
            if p in matches:
                return None

        for tactic in self.tactics():
            if tactic.trigger(entity.relpath(bd)):
                return tactic(target=target, entity=entity,
                              current=current, config=next_config)
        return None



#!/usr/bin/env python

import sys

import common
from charmhelpers.core import hookenv, host


hooks = hookenv.Hooks()
mountpoint = '/srv/mysql'


@hooks.hook('data-relation-joined', 'data-relation-changed')
def data_relation():
    if hookenv.relation_get('mountpoint') == mountpoint:
        # Other side of relation is ready
        common.migrate_to_mount(mountpoint)
    else:
        # Other side not ready yet, provide mountpoint
        hookenv.log('Requesting storage for {}'.format(mountpoint))
        hookenv.relation_set(mountpoint=mountpoint)


@hooks.hook('data-relation-departed', 'data-relation-broken')
def data_relation_gone():
    hookenv.log('Data relation no longer present, stopping MysQL.')
    host.service_stop('mysql')


if __name__ == '__main__':
    hooks.execute(sys.argv)



# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Helper functions for writing Juju charms in Python."""

__metaclass__ = type
__all__ = ['get_config',
           'log',
           'log_entry',
           'log_exit',
           'relation_get',
           'relation_set',
           'relation_ids',
           'relation_list',
           'config_get',
           'unit_get',
           'open_port',
           'close_port',
           'service_control',
           'unit_info',
           'wait_for_machine',
           'wait_for_page_contents',
           'wait_for_relation',
           'wait_for_unit',
           ]

from collections import namedtuple
import json
import operator
from shelltoolbox import (
    command,
    script_name,
    run,
)
import tempfile
import time
import urllib2
import yaml
from subprocess import CalledProcessError


SLEEP_AMOUNT = 0.1
Env = namedtuple('Env', 'uid gid home')
# We create a juju_status Command here because it makes testing much,
# much easier.
juju_status = lambda: command('juju')('status')


def log(message, juju_log=command('juju-log')):
    return juju_log('--', message)


def log_entry():
    log("--> Entering {}".format(script_name()))


def log_exit():
    log("<-- Exiting {}".format(script_name()))


def get_config():
    _config_get = command('config-get', '--format=json')
    return json.loads(_config_get())


def relation_get(attribute=None, unit=None, rid=None):
    cmd = command('relation-get')
    if attribute is None and unit is None and rid is None:
        return cmd().strip()
    _args = []
    if rid:
        _args.append('-r')
        _args.append(rid)
    if attribute is not None:
        _args.append(attribute)
    if unit:
        _args.append(unit)
    return cmd(*_args).strip()


def relation_set(**kwargs):
    cmd = command('relation-set')
    args = ['{}={}'.format(k, v) for k, v in kwargs.items()]
    cmd(*args)


def relation_ids(relation_name):
    cmd = command('relation-ids')
    args = [relation_name]
    return cmd(*args).split()


def relation_list(rid=None):
    cmd = command('relation-list')
    args = []
    if rid:
        args.append('-r')
        args.append(rid)
    return cmd(*args).split()


def config_get(attribute):
    cmd = command('config-get')
    args = [attribute]
    return cmd(*args).strip()


def unit_get(attribute):
    cmd = command('unit-get')
    args = [attribute]
    return cmd(*args).strip()


def open_port(port, protocol="TCP"):
    cmd = command('open-port')
    args = ['{}/{}'.format(port, protocol)]
    cmd(*args)


def close_port(port, protocol="TCP"):
    cmd = command('close-port')
    args = ['{}/{}'.format(port, protocol)]
    cmd(*args)

START = "start"
RESTART = "restart"
STOP = "stop"
RELOAD = "reload"


def service_control(service_name, action):
    cmd = command('service')
    args = [service_name, action]
    try:
        if action == RESTART:
            try:
                cmd(*args)
            except CalledProcessError:
                service_control(service_name, START)
        else:
            cmd(*args)
    except CalledProcessError:
        log("Failed to perform {} on service {}".format(action, service_name))


def configure_source(update=False):
    source = config_get('source')
    if ((source.startswith('ppa:') or
         source.startswith('cloud:') or
         source.startswith('http:'))):
        run('add-apt-repository', source)
    if source.startswith("http:"):
        run('apt-key', 'import', config_get('key'))
    if update:
        run('apt-get', 'update')


def make_charm_config_file(charm_config):
    charm_config_file = tempfile.NamedTemporaryFile()
    charm_config_file.write(yaml.dump(charm_config))
    charm_config_file.flush()
    # The NamedTemporaryFile instance is returned instead of just the name
    # because we want to take advantage of garbage collection-triggered
    # deletion of the temp file when it goes out of scope in the caller.
    return charm_config_file


def unit_info(service_name, item_name, data=None, unit=None):
    if data is None:
        data = yaml.safe_load(juju_status())
    service = data['services'].get(service_name)
    if service is None:
        # XXX 2012-02-08 gmb:
        #     This allows us to cope with the race condition that we
        #     have between deploying a service and having it come up in
        #     `juju status`. We could probably do with cleaning it up so
        #     that it fails a bit more noisily after a while.
        return ''
    units = service['units']
    if unit is not None:
        item = units[unit][item_name]
    else:
        # It might seem odd to sort the units here, but we do it to
        # ensure that when no unit is specified, the first unit for the
        # service (or at least the one with the lowest number) is the
        # one whose data gets returned.
        sorted_unit_names = sorted(units.keys())
        item = units[sorted_unit_names[0]][item_name]
    return item


def get_machine_data():
    return yaml.safe_load(juju_status())['machines']


def wait_for_machine(num_machines=1, timeout=300):
    """Wait `timeout` seconds for `num_machines` machines to come up.

    This wait_for... function can be called by other wait_for functions
    whose timeouts might be too short in situations where only a bare
    Juju setup has been bootstrapped.

    :return: A tuple of (num_machines, time_taken). This is used for
             testing.
    """
    # You may think this is a hack, and you'd be right. The easiest way
    # to tell what environment we're working in (LXC vs EC2) is to check
    # the dns-name of the first machine. If it's localhost we're in LXC
    # and we can just return here.
    if get_machine_data()[0]['dns-name'] == 'localhost':
        return 1, 0
    start_time = time.time()
    while True:
        # Drop the first machine, since it's the Zookeeper and that's
        # not a machine that we need to wait for. This will only work
        # for EC2 environments, which is why we return early above if
        # we're in LXC.
        machine_data = get_machine_data()
        non_zookeeper_machines = [
            machine_data[key] for key in machine_data.keys()[1:]]
        if len(non_zookeeper_machines) >= num_machines:
            all_machines_running = True
            for machine in non_zookeeper_machines:
                if machine.get('instance-state') != 'running':
                    all_machines_running = False
                    break
            if all_machines_running:
                break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for service to start')
        time.sleep(SLEEP_AMOUNT)
    return num_machines, time.time() - start_time


def wait_for_unit(service_name, timeout=480):
    """Wait `timeout` seconds for a given service name to come up."""
    wait_for_machine(num_machines=1)
    start_time = time.time()
    while True:
        state = unit_info(service_name, 'agent-state')
        if 'error' in state or state == 'started':
            break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for service to start')
        time.sleep(SLEEP_AMOUNT)
    if state != 'started':
        raise RuntimeError('unit did not start, agent-state: ' + state)


def wait_for_relation(service_name, relation_name, timeout=120):
    """Wait `timeout` seconds for a given relation to come up."""
    start_time = time.time()
    while True:
        relation = unit_info(service_name, 'relations').get(relation_name)
        if relation is not None and relation['state'] == 'up':
            break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for relation to be up')
        time.sleep(SLEEP_AMOUNT)


def wait_for_page_contents(url, contents, timeout=120, validate=None):
    if validate is None:
        validate = operator.contains
    start_time = time.time()
    while True:
        try:
            stream = urllib2.urlopen(url)
        except (urllib2.HTTPError, urllib2.URLError):
            pass
        else:
            page = stream.read()
            if validate(page, contents):
                return page
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for contents of ' + url)
        time.sleep(SLEEP_AMOUNT)



import os

import requests
from charmtools import fetchers
from charmtools.fetchers import (git,  # noqa
                                 Fetcher,
                                 get_fetcher,
                                 FetchError)

from path import path


class RepoFetcher(fetchers.LocalFetcher):
    @classmethod
    def can_fetch(cls, url):
        search_path = [os.getcwd(), os.environ.get("JUJU_REPOSITORY", ".")]
        cp = os.environ.get("LAYER_PATH")
        if cp:
            search_path.extend(cp.split(":"))
        for part in search_path:
            p = (path(part) / url).normpath()
            if p.exists():
                return dict(path=p)
        return {}

fetchers.FETCHERS.insert(0, RepoFetcher)


class InterfaceFetcher(fetchers.LocalFetcher):
    # XXX: When hosted somewhere, fix this
    INTERFACE_DOMAIN = "http://interfaces.juju.solutions"
    NAMESPACE = "interface"
    ENVIRON = "INTERFACE_PATH"
    OPTIONAL_PREFIX = "juju-relation-"
    ENDPOINT = "/api/v1/interface"
    NO_LOCAL_LAYERS = False

    @classmethod
    def can_fetch(cls, url):
        # Search local path first, then the interface webservice
        if url.startswith("{}:".format(cls.NAMESPACE)):
            name = url[len(cls.NAMESPACE) + 1:]

            if not cls.NO_LOCAL_LAYERS:
                prefixed_name = '{}-{}'.format(cls.NAMESPACE, name)
                search_path = [os.environ.get("JUJU_REPOSITORY", ".")]
                cp = os.environ.get(cls.ENVIRON)
                if cp:
                    search_path.extend(cp.split(os.pathsep))
                for part in search_path:
                    basepath = path(part)
                    for dirname in (name, prefixed_name):
                        p = (basepath / dirname).normpath()
                        if p.exists():
                            return dict(path=p)

            choices = [name]
            if name.startswith(cls.OPTIONAL_PREFIX):
                choices.append(name[len(cls.OPTIONAL_PREFIX):])
            for choice in choices:
                uri = "%s%s/%s/" % (
                    cls.INTERFACE_DOMAIN, cls.ENDPOINT, choice)
                try:
                    result = requests.get(uri)
                except:
                    result = None
                if result and result.ok:
                    result = result.json()
                    if result.get("repo"):
                        return result
            return {}

    def target(self, dir_):
        """Return a :class:`path` of the directory where the downloaded item
        will be located.

        :param str dir_: Directory into which the item will be downloaded.
        :return: :class:`path`

        """
        if hasattr(self, "path"):
            return self.path
        elif hasattr(self, "repo"):
            _, target = self._get_repo_fetcher_and_target(self.repo, dir_)
            return target

    def _get_repo_fetcher_and_target(self, repo, dir_):
        """Returns a :class:`Fetcher` for ``repo``, and the destination dir
        at which the downloaded repo will be created.

        :param str repo: The repo url.
        :param str dir_: Directory into which the repo will be downloaded.
        :return: 2-tuple of (:class:`Fetcher`, :class:`path`)

        """
        u = self.url[len(self.NAMESPACE) + 1:]
        f = get_fetcher(repo)
        if hasattr(f, "repo"):
            basename = path(f.repo).name.splitext()[0]
        else:
            basename = u
        return f, path(dir_) / basename

    def fetch(self, dir_):
        if hasattr(self, "path"):
            return super(InterfaceFetcher, self).fetch(dir_)
        elif hasattr(self, "repo"):
            f, target = self._get_repo_fetcher_and_target(self.repo, dir_)
            res = f.fetch(dir_)
            if res != target:
                target.rmtree_p()
                path(res).rename(target)
            return target

fetchers.FETCHERS.insert(0, InterfaceFetcher)


class LayerFetcher(InterfaceFetcher):
    INTERFACE_DOMAIN = "http://interfaces.juju.solutions"
    NAMESPACE = "layer"
    ENVIRON = "LAYER_PATH"
    OPTIONAL_PREFIX = "juju-layer-"
    ENDPOINT = "/api/v1/layer"

fetchers.FETCHERS.insert(0, LayerFetcher)



# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import six
from charmhelpers.core.hookenv import relation_id as current_relation_id
from charmhelpers.core.hookenv import (
    is_relation_made,
    relation_ids,
    relation_get,
    local_unit,
    relation_set,
)


"""
This helper provides functions to support use of a peer relation
for basic key/value storage, with the added benefit that all storage
can be replicated across peer units.

Requirement to use:

To use this, the "peer_echo()" method has to be called form the peer
relation's relation-changed hook:

@hooks.hook("cluster-relation-changed") # Adapt the to your peer relation name
def cluster_relation_changed():
    peer_echo()

Once this is done, you can use peer storage from anywhere:

@hooks.hook("some-hook")
def some_hook():
    # You can store and retrieve key/values this way:
    if is_relation_made("cluster"):  # from charmhelpers.core.hookenv
        # There are peers available so we can work with peer storage
        peer_store("mykey", "myvalue")
        value = peer_retrieve("mykey")
        print value
    else:
        print "No peers joind the relation, cannot share key/values :("
"""


def peer_retrieve(key, relation_name='cluster'):
    """Retrieve a named key from peer relation `relation_name`."""
    cluster_rels = relation_ids(relation_name)
    if len(cluster_rels) > 0:
        cluster_rid = cluster_rels[0]
        return relation_get(attribute=key, rid=cluster_rid,
                            unit=local_unit())
    else:
        raise ValueError('Unable to detect'
                         'peer relation {}'.format(relation_name))


def peer_retrieve_by_prefix(prefix, relation_name='cluster', delimiter='_',
                            inc_list=None, exc_list=None):
    """ Retrieve k/v pairs given a prefix and filter using {inc,exc}_list """
    inc_list = inc_list if inc_list else []
    exc_list = exc_list if exc_list else []
    peerdb_settings = peer_retrieve('-', relation_name=relation_name)
    matched = {}
    for k, v in peerdb_settings.items():
        full_prefix = prefix + delimiter
        if k.startswith(full_prefix):
            new_key = k.replace(full_prefix, '')
            if new_key in exc_list:
                continue
            if new_key in inc_list or len(inc_list) == 0:
                matched[new_key] = v
    return matched


def peer_store(key, value, relation_name='cluster'):
    """Store the key/value pair on the named peer relation `relation_name`."""
    cluster_rels = relation_ids(relation_name)
    if len(cluster_rels) > 0:
        cluster_rid = cluster_rels[0]
        relation_set(relation_id=cluster_rid,
                     relation_settings={key: value})
    else:
        raise ValueError('Unable to detect '
                         'peer relation {}'.format(relation_name))


def peer_echo(includes=None):
    """Echo filtered attributes back onto the same relation for storage.

    This is a requirement to use the peerstorage module - it needs to be called
    from the peer relation's changed hook.
    """
    rdata = relation_get()
    echo_data = {}
    if includes is None:
        echo_data = rdata.copy()
        for ex in ['private-address', 'public-address']:
            if ex in echo_data:
                echo_data.pop(ex)
    else:
        for attribute, value in six.iteritems(rdata):
            for include in includes:
                if include in attribute:
                    echo_data[attribute] = value
    if len(echo_data) > 0:
        relation_set(relation_settings=echo_data)


def peer_store_and_set(relation_id=None, peer_relation_name='cluster',
                       peer_store_fatal=False, relation_settings=None,
                       delimiter='_', **kwargs):
    """Store passed-in arguments both in argument relation and in peer storage.

    It functions like doing relation_set() and peer_store() at the same time,
    with the same data.

    @param relation_id: the id of the relation to store the data on. Defaults
                        to the current relation.
    @param peer_store_fatal: Set to True, the function will raise an exception
                             should the peer sotrage not be avialable."""

    relation_settings = relation_settings if relation_settings else {}
    relation_set(relation_id=relation_id,
                 relation_settings=relation_settings,
                 **kwargs)
    if is_relation_made(peer_relation_name):
        for key, value in six.iteritems(dict(list(kwargs.items()) +
                                             list(relation_settings.items()))):
            key_prefix = relation_id or current_relation_id()
            peer_store(key_prefix + delimiter + key,
                       value,
                       relation_name=peer_relation_name)
    else:
        if peer_store_fatal:
            raise ValueError('Unable to detect '
                             'peer relation {}'.format(peer_relation_name))




import os
import pkg_resources
import argparse

from cli import parser_defaults
from charmtools import utils


def get_args(args=None):
    parser = argparse.ArgumentParser(
        description='display tooling version information')
    utils.add_plugin_description(parser)
    parser = parser_defaults(parser)
    args = parser.parse_args(args)

    return args


def charm_version():
    if 'SNAP' in os.environ:
        cscv = os.path.join(os.environ['SNAP'], 'charmstore-client-version')
        if os.path.exists(cscv):
            with open(cscv) as f:
                charm_ver = f.read().strip()
            return charm_ver
    try:
        from apt.cache import Cache
        charm_vers = Cache()['charm'].versions
        for v in charm_vers:
            if v.is_installed:
                charm_ver = v.version
                break
    except ImportError:
        charm_ver = 'unavailable'
    except:
        charm_ver = 'error'

    return charm_ver


def main():
    get_args()

    version = pkg_resources.get_distribution("charm-tools").version

    print "charm %s" % charm_version()
    print "charm-tools %s" % version


if __name__ == '__main__':
    main()



"""Unit test for juju tests substrate handling"""

import unittest
import yaml
from charmtools.test import parse_substrates, allowed_substrates


class JujuTestSubstrateTests(unittest.TestCase):
    INCLUDE_SKIP = '''
substrates:
    order: include, skip
    skip: amazon
    include: "*"
    '''

    SKIP_INCLUDE = '''
substrates:
    order: skip, include
    skip: "*"
    include: amazon
    '''

    def test_parse_substrates_text(self):
        self.assertRaises(ValueError,  parse_substrates, '')
        self.assertIsNotNone(parse_substrates(self.INCLUDE_SKIP))

    def test_parse_substrate_dict(self):
        data = yaml.load(self.INCLUDE_SKIP)
        self.assertIsNotNone(parse_substrates(data))

    def test_parse_substrate_object(self):
        class Object(object):
            pass
        data = yaml.load(self.INCLUDE_SKIP)
        o = Object()
        o.__dict__ = data
        self.assertIsNotNone(parse_substrates(o))

    def test_parse_substrates_order(self):
        result = parse_substrates(self.INCLUDE_SKIP)
        self.assertEqual(result.order, ['include', 'skip'])

    def test_parse_substrates_order_invalid(self):
        data = self.INCLUDE_SKIP.replace('skip', 'foobar')
        self.assertRaises(ValueError, parse_substrates, data)

    def test_parse_substrates_include(self):
        result = parse_substrates(self.INCLUDE_SKIP)
        self.assertEqual(result.include, set(['*']))

    def test_parse_substrates_skip(self):
        result = parse_substrates(self.INCLUDE_SKIP)
        self.assertEqual(result.skip, set(['amazon']))

    def test_filter_include_skip(self):
        rules = parse_substrates(self.INCLUDE_SKIP)
        self.assertEqual(rules.filter('amazon,hp,azure'), ['azure', 'hp'])

    def test_filter_skip_include(self):
        rules = parse_substrates(self.SKIP_INCLUDE)
        self.assertEqual(rules.filter('amazon,hp,azure'), ['amazon'])

    def test_allowed_substrates(self):
        self.assertEqual(allowed_substrates(self.INCLUDE_SKIP,
                                            ['foo', 'bar', 'amazon']),
                         ['bar', 'foo'])


if __name__ == '__main__':
    unittest.main()



import unittest

try:
    import mock
except ImportError:
    raise ImportError(
        "Please ensure both python-mock and python-nose are installed.")


from hooks import hooks


class InstallHookTestCase(unittest.TestCase):

    def setUp(self):
        super(InstallHookTestCase, self).setUp()

        patcher = mock.patch('hooks.charmhelpers')
        self.mock_charmhelpers = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_charmhelpers.core.hookenv.config.return_value = {
            'install_deps_from_ppa': False,
        }

        patcher = mock.patch('charmhelpers.contrib.ansible.apply_playbook')
        self.mock_apply_playbook = patcher.start()
        self.addCleanup(patcher.stop)

    def test_installs_ansible_support(self):
        hooks.execute(['install'])

        ansible = self.mock_charmhelpers.contrib.ansible
        ansible.install_ansible_support.assert_called_once_with(
            from_ppa=True)

    def test_applies_install_playbook(self):
        hooks.execute(['install'])

        self.assertEqual([
            mock.call('playbooks/site.yaml', tags=['install']),
        ], self.mock_apply_playbook.call_args_list)


class DefaultHooksTestCase(unittest.TestCase):

    def setUp(self):
        super(DefaultHooksTestCase, self).setUp()
        patcher = mock.patch('charmhelpers.contrib.ansible.apply_playbook')
        self.mock_apply_playbook = patcher.start()
        self.addCleanup(patcher.stop)

    def test_default_hooks(self):
        """Most of the hooks let ansible do all the work."""
        for hook in ('start', 'stop', 'config-changed'):
            self.mock_apply_playbook.reset_mock()

            hooks.execute([hook])

            self.assertEqual([
                mock.call('playbooks/site.yaml',
                          tags=[hook]),
            ], self.mock_apply_playbook.call_args_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Authors:
#  Kapil Thangavelu <kapil.foss@gmail.com>
#
"""
Intro
-----

A simple way to store state in units. This provides a key value
storage with support for versioned, transactional operation,
and can calculate deltas from previous values to simplify unit logic
when processing changes.


Hook Integration
----------------

There are several extant frameworks for hook execution, including

 - charmhelpers.core.hookenv.Hooks
 - charmhelpers.core.services.ServiceManager

The storage classes are framework agnostic, one simple integration is
via the HookData contextmanager. It will record the current hook
execution environment (including relation data, config data, etc.),
setup a transaction and allow easy access to the changes from
previously seen values. One consequence of the integration is the
reservation of particular keys ('rels', 'unit', 'env', 'config',
'charm_revisions') for their respective values.

Here's a fully worked integration example using hookenv.Hooks::

       from charmhelper.core import hookenv, unitdata

       hook_data = unitdata.HookData()
       db = unitdata.kv()
       hooks = hookenv.Hooks()

       @hooks.hook
       def config_changed():
           # Print all changes to configuration from previously seen
           # values.
           for changed, (prev, cur) in hook_data.conf.items():
               print('config changed', changed,
                     'previous value', prev,
                     'current value',  cur)

           # Get some unit specific bookeeping
           if not db.get('pkg_key'):
               key = urllib.urlopen('https://example.com/pkg_key').read()
               db.set('pkg_key', key)

           # Directly access all charm config as a mapping.
           conf = db.getrange('config', True)

           # Directly access all relation data as a mapping
           rels = db.getrange('rels', True)

       if __name__ == '__main__':
           with hook_data():
               hook.execute()


A more basic integration is via the hook_scope context manager which simply
manages transaction scope (and records hook name, and timestamp)::

  >>> from unitdata import kv
  >>> db = kv()
  >>> with db.hook_scope('install'):
  ...    # do work, in transactional scope.
  ...    db.set('x', 1)
  >>> db.get('x')
  1


Usage
-----

Values are automatically json de/serialized to preserve basic typing
and complex data struct capabilities (dicts, lists, ints, booleans, etc).

Individual values can be manipulated via get/set::

   >>> kv.set('y', True)
   >>> kv.get('y')
   True

   # We can set complex values (dicts, lists) as a single key.
   >>> kv.set('config', {'a': 1, 'b': True'})

   # Also supports returning dictionaries as a record which
   # provides attribute access.
   >>> config = kv.get('config', record=True)
   >>> config.b
   True


Groups of keys can be manipulated with update/getrange::

   >>> kv.update({'z': 1, 'y': 2}, prefix="gui.")
   >>> kv.getrange('gui.', strip=True)
   {'z': 1, 'y': 2}

When updating values, its very helpful to understand which values
have actually changed and how have they changed. The storage
provides a delta method to provide for this::

   >>> data = {'debug': True, 'option': 2}
   >>> delta = kv.delta(data, 'config.')
   >>> delta.debug.previous
   None
   >>> delta.debug.current
   True
   >>> delta
   {'debug': (None, True), 'option': (None, 2)}

Note the delta method does not persist the actual change, it needs to
be explicitly saved via 'update' method::

   >>> kv.update(data, 'config.')

Values modified in the context of a hook scope retain historical values
associated to the hookname.

   >>> with db.hook_scope('config-changed'):
   ...      db.set('x', 42)
   >>> db.gethistory('x')
   [(1, u'x', 1, u'install', u'2015-01-21T16:49:30.038372'),
    (2, u'x', 42, u'config-changed', u'2015-01-21T16:49:30.038786')]

"""

import collections
import contextlib
import datetime
import json
import os
import pprint
import sqlite3
import sys

__author__ = 'Kapil Thangavelu <kapil.foss@gmail.com>'


class Storage(object):
    """Simple key value database for local unit state within charms.

    Modifications are automatically committed at hook exit. That's
    currently regardless of exit code.

    To support dicts, lists, integer, floats, and booleans values
    are automatically json encoded/decoded.
    """
    def __init__(self, path=None):
        self.db_path = path
        if path is None:
            self.db_path = os.path.join(
                os.environ.get('CHARM_DIR', ''), '.unit-state.db')
        self.conn = sqlite3.connect('%s' % self.db_path)
        self.cursor = self.conn.cursor()
        self.revision = None
        self._closed = False
        self._init()

    def close(self):
        if self._closed:
            return
        self.flush(False)
        self.cursor.close()
        self.conn.close()
        self._closed = True

    def _scoped_query(self, stmt, params=None):
        if params is None:
            params = []
        return stmt, params

    def get(self, key, default=None, record=False):
        self.cursor.execute(
            *self._scoped_query(
                'select data from kv where key=?', [key]))
        result = self.cursor.fetchone()
        if not result:
            return default
        if record:
            return Record(json.loads(result[0]))
        return json.loads(result[0])

    def getrange(self, key_prefix, strip=False):
        stmt = "select key, data from kv where key like '%s%%'" % key_prefix
        self.cursor.execute(*self._scoped_query(stmt))
        result = self.cursor.fetchall()

        if not result:
            return None
        if not strip:
            key_prefix = ''
        return dict([
            (k[len(key_prefix):], json.loads(v)) for k, v in result])

    def update(self, mapping, prefix=""):
        for k, v in mapping.items():
            self.set("%s%s" % (prefix, k), v)

    def unset(self, key):
        self.cursor.execute('delete from kv where key=?', [key])
        if self.revision and self.cursor.rowcount:
            self.cursor.execute(
                'insert into kv_revisions values (?, ?, ?)',
                [key, self.revision, json.dumps('DELETED')])

    def set(self, key, value):
        serialized = json.dumps(value)

        self.cursor.execute(
            'select data from kv where key=?', [key])
        exists = self.cursor.fetchone()

        # Skip mutations to the same value
        if exists:
            if exists[0] == serialized:
                return value

        if not exists:
            self.cursor.execute(
                'insert into kv (key, data) values (?, ?)',
                (key, serialized))
        else:
            self.cursor.execute('''
            update kv
            set data = ?
            where key = ?''', [serialized, key])

        # Save
        if not self.revision:
            return value

        self.cursor.execute(
            'select 1 from kv_revisions where key=? and revision=?',
            [key, self.revision])
        exists = self.cursor.fetchone()

        if not exists:
            self.cursor.execute(
                '''insert into kv_revisions (
                revision, key, data) values (?, ?, ?)''',
                (self.revision, key, serialized))
        else:
            self.cursor.execute(
                '''
                update kv_revisions
                set data = ?
                where key = ?
                and   revision = ?''',
                [serialized, key, self.revision])

        return value

    def delta(self, mapping, prefix):
        """
        return a delta containing values that have changed.
        """
        previous = self.getrange(prefix, strip=True)
        if not previous:
            pk = set()
        else:
            pk = set(previous.keys())
        ck = set(mapping.keys())
        delta = DeltaSet()

        # added
        for k in ck.difference(pk):
            delta[k] = Delta(None, mapping[k])

        # removed
        for k in pk.difference(ck):
            delta[k] = Delta(previous[k], None)

        # changed
        for k in pk.intersection(ck):
            c = mapping[k]
            p = previous[k]
            if c != p:
                delta[k] = Delta(p, c)

        return delta

    @contextlib.contextmanager
    def hook_scope(self, name=""):
        """Scope all future interactions to the current hook execution
        revision."""
        assert not self.revision
        self.cursor.execute(
            'insert into hooks (hook, date) values (?, ?)',
            (name or sys.argv[0],
             datetime.datetime.utcnow().isoformat()))
        self.revision = self.cursor.lastrowid
        try:
            yield self.revision
            self.revision = None
        except:
            self.flush(False)
            self.revision = None
            raise
        else:
            self.flush()

    def flush(self, save=True):
        if save:
            self.conn.commit()
        elif self._closed:
            return
        else:
            self.conn.rollback()

    def _init(self):
        self.cursor.execute('''
            create table if not exists kv (
               key text,
               data text,
               primary key (key)
               )''')
        self.cursor.execute('''
            create table if not exists kv_revisions (
               key text,
               revision integer,
               data text,
               primary key (key, revision)
               )''')
        self.cursor.execute('''
            create table if not exists hooks (
               version integer primary key autoincrement,
               hook text,
               date text
               )''')
        self.conn.commit()

    def gethistory(self, key, deserialize=False):
        self.cursor.execute(
            '''
            select kv.revision, kv.key, kv.data, h.hook, h.date
            from kv_revisions kv,
                 hooks h
            where kv.key=?
             and kv.revision = h.version
            ''', [key])
        if deserialize is False:
            return self.cursor.fetchall()
        return map(_parse_history, self.cursor.fetchall())

    def debug(self, fh=sys.stderr):
        self.cursor.execute('select * from kv')
        pprint.pprint(self.cursor.fetchall(), stream=fh)
        self.cursor.execute('select * from kv_revisions')
        pprint.pprint(self.cursor.fetchall(), stream=fh)


def _parse_history(d):
    return (d[0], d[1], json.loads(d[2]), d[3],
            datetime.datetime.strptime(d[-1], "%Y-%m-%dT%H:%M:%S.%f"))


class HookData(object):
    """Simple integration for existing hook exec frameworks.

    Records all unit information, and stores deltas for processing
    by the hook.

    Sample::

       from charmhelper.core import hookenv, unitdata

       changes = unitdata.HookData()
       db = unitdata.kv()
       hooks = hookenv.Hooks()

       @hooks.hook
       def config_changed():
           # View all changes to configuration
           for changed, (prev, cur) in changes.conf.items():
               print('config changed', changed,
                     'previous value', prev,
                     'current value',  cur)

           # Get some unit specific bookeeping
           if not db.get('pkg_key'):
               key = urllib.urlopen('https://example.com/pkg_key').read()
               db.set('pkg_key', key)

       if __name__ == '__main__':
           with changes():
               hook.execute()

    """
    def __init__(self):
        self.kv = kv()
        self.conf = None
        self.rels = None

    @contextlib.contextmanager
    def __call__(self):
        from charmhelpers.core import hookenv
        hook_name = hookenv.hook_name()

        with self.kv.hook_scope(hook_name):
            self._record_charm_version(hookenv.charm_dir())
            delta_config, delta_relation = self._record_hook(hookenv)
            yield self.kv, delta_config, delta_relation

    def _record_charm_version(self, charm_dir):
        # Record revisions.. charm revisions are meaningless
        # to charm authors as they don't control the revision.
        # so logic dependnent on revision is not particularly
        # useful, however it is useful for debugging analysis.
        charm_rev = open(
            os.path.join(charm_dir, 'revision')).read().strip()
        charm_rev = charm_rev or '0'
        revs = self.kv.get('charm_revisions', [])
        if charm_rev not in revs:
            revs.append(charm_rev.strip() or '0')
            self.kv.set('charm_revisions', revs)

    def _record_hook(self, hookenv):
        data = hookenv.execution_environment()
        self.conf = conf_delta = self.kv.delta(data['conf'], 'config')
        self.rels = rels_delta = self.kv.delta(data['rels'], 'rels')
        self.kv.set('env', data['env'])
        self.kv.set('unit', data['unit'])
        self.kv.set('relid', data.get('relid'))
        return conf_delta, rels_delta


class Record(dict):

    __slots__ = ()

    def __getattr__(self, k):
        if k in self:
            return self[k]
        raise AttributeError(k)


class DeltaSet(Record):

    __slots__ = ()


Delta = collections.namedtuple('Delta', ['previous', 'current'])


_KV = None


def kv():
    global _KV
    if _KV is None:
        _KV = Storage()
    return _KV



# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import os
from charmhelpers.fetch import (
    BaseFetchHandler,
    UnhandledSource
)
from charmhelpers.core.host import mkdir

import six
if six.PY3:
    raise ImportError('bzrlib does not support Python3')

try:
    from bzrlib.branch import Branch
    from bzrlib import bzrdir, workingtree, errors
except ImportError:
    from charmhelpers.fetch import apt_install
    apt_install("python-bzrlib")
    from bzrlib.branch import Branch
    from bzrlib import bzrdir, workingtree, errors


class BzrUrlFetchHandler(BaseFetchHandler):
    """Handler for bazaar branches via generic and lp URLs"""
    def can_handle(self, source):
        url_parts = self.parse_url(source)
        if url_parts.scheme not in ('bzr+ssh', 'lp'):
            return False
        else:
            return True

    def branch(self, source, dest):
        url_parts = self.parse_url(source)
        # If we use lp:branchname scheme we need to load plugins
        if not self.can_handle(source):
            raise UnhandledSource("Cannot handle {}".format(source))
        if url_parts.scheme == "lp":
            from bzrlib.plugin import load_plugins
            load_plugins()
        try:
            local_branch = bzrdir.BzrDir.create_branch_convenience(dest)
        except errors.AlreadyControlDirError:
            local_branch = Branch.open(dest)
        try:
            remote_branch = Branch.open(source)
            remote_branch.push(local_branch)
            tree = workingtree.WorkingTree.open(dest)
            tree.update()
        except Exception as e:
            raise e

    def install(self, source):
        url_parts = self.parse_url(source)
        branch_name = url_parts.path.strip("/").split("/")[-1]
        dest_dir = os.path.join(os.environ.get('CHARM_DIR'), "fetched",
                                branch_name)
        if not os.path.exists(dest_dir):
            mkdir(dest_dir, perms=0o755)
        try:
            self.branch(source, dest_dir)
        except OSError as e:
            raise UnhandledSource(e.strerror)
        return dest_dir



"""Unit test for juju_test"""

import os
import unittest
import yaml

from contextlib import contextmanager
from charmtools import test as juju_test
from mock import patch, call, Mock, MagicMock

RAW_ENVIRONMENTS_YAML = '''
default: gojuju
environments:
  gojuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: gojuju-xyxyz
    admin-secret: zyxyx
    default-series: world
  pyjuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: juju-xyxyz
    admin-secret: zyxyxy
    default-series: world
    juju-origin: ppa
    ssl-hostname-verification: true'''

PARSED_ENVIRONMENTS_YAML = yaml.safe_load(RAW_ENVIRONMENTS_YAML)


@contextmanager
def cd(directory):
    """A context manager to temporarily change current working dir, e.g.::

        >>> import os
        >>> os.chdir('/tmp')
        >>> with cd('/bin'): print os.getcwd()
        /bin
        >>> print os.getcwd()
        /tmp
    """
    cwd = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(cwd)


class Arguments(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, name):
        return None


class JujuTestPluginTest(unittest.TestCase):
    @patch('subprocess.check_output')
    def test_get_gojuju_version(self, mock_check_output):
        mock_check_output.side_effect = ['1.2.3-series-xxx']
        version = juju_test.get_juju_version()

        self.assertEqual(version.major, 1)
        self.assertEqual(version.minor, 2)
        self.assertEqual(version.patch, 3)

        mock_check_output.assert_called_with(['juju', 'version'])

    @patch('subprocess.check_output')
    def test_get_pyjuju_version(self, mcheck_output):
        mcheck_output.side_effect = [Exception('Non-zero exit'), 'juju 8.6']
        version = juju_test.get_juju_version()

        self.assertEqual(version.major, 8)
        self.assertEqual(version.minor, 6)
        self.assertEqual(version.patch, 0)

        mcheck_output.assert_called_with(['juju', '--version'])

    @patch('subprocess.check_output')
    def test_get_juju_version_malformed(self, mcheck_output):
        mcheck_output.return_value = '1.2.3.45'
        version = juju_test.get_juju_version()

        self.assertEqual(version.major, 1)
        self.assertEqual(version.minor, 2)
        self.assertEqual(version.patch, 3)

        mcheck_output.assert_called_once_with(['juju', 'version'])

    def test_convert_to_timedelta(self):
        self.assertEqual(300, juju_test.convert_to_timedelta('5m'))
        self.assertEqual(300, juju_test.convert_to_timedelta('300'))
        self.assertEqual(100, juju_test.convert_to_timedelta('100s'))
        self.assertEqual(100, juju_test.convert_to_timedelta(100))
        self.assertEqual(60, juju_test.convert_to_timedelta('1m'))
        self.assertEqual(60 * 60, juju_test.convert_to_timedelta('1h'))

    def test_conductor_find_tests(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(tests="dummy")
            c = juju_test.Conductor(args)
            results = c.find_tests()
            self.assertIn('00_setup', results)
            self.assertNotIn('helpers.bash', results)

    def test_conductor_find_tests_fails(self):
        with cd('tests_functional/charms/mod-spdy/'):
            args = Arguments(tests="dummy")
            self.assertRaises(juju_test.NoTests, juju_test.Conductor, args)

    def test_conductor_ignores_arbitrary_env(self):
        """Ensure that the conductor ignores other env vars."""
        os.environ['CHARMTOOLS_TEST'] = 'FOOBAR'
        args = Arguments(tests="dummy")
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            self.assertNotIn('CHARMTOOLS_TEST', c.env)

    def test_conductor_keeps_whitelist_env(self):
        """Ensure that the conductor copies the environment whitelist."""
        os.environ['PATH'] = 'FOOBAR'
        os.environ['SSH_AGENT_PID'] = 'FOOBAR'
        os.environ['SSH_AUTH_SOCK'] = 'FOOBAR'
        args = Arguments(tests="dummy")
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            self.assertEqual('FOOBAR', c.env['PATH'])
            self.assertEqual('FOOBAR', c.env['SSH_AGENT_PID'])
            self.assertEqual('FOOBAR', c.env['SSH_AUTH_SOCK'])

    def test_conductor_allows_addition_to_whitelist_env(self):
        """Ensure that the conductor copies the environment whitelist."""
        with cd('tests_functional/charms/test/'):
            os.environ['FOO'] = 'BAR'
            os.environ['BAR'] = 'BAZ'
            args = Arguments(tests="dummy")
            args.preserve_environment_variables = "FOO,BAR"
            c = juju_test.Conductor(args)
            self.assertEqual('BAR', c.env['FOO'])
            self.assertEqual('BAZ', c.env['BAR'])

    @patch.object(juju_test.Conductor, 'find_tests')
    def test_conductor_find_tests_exception(self, mfind_tests):
        mfind_tests.return_value = None
        args = Arguments(tests=None)
        self.assertRaises(juju_test.NoTests, juju_test.Conductor, args)

    @patch('subprocess.check_output')
    def test_conductor_status(self, mcheck_output):
        yml_output = '''
        machines:
            '0':
                agent-state: limbo
                agent-version: 999.999.999
                dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
                instance-id: j-xxxxxx
                series: world
        services: {}'''

        mcheck_output.side_effect = [yml_output, Exception('not bootstrapped')]
        juju_env = 'test'
        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            results = c.status(juju_env)

            mcheck_output.assert_called_with(
                ['juju', 'status', '-e', juju_env], env=c.env)
            self.assertEqual(results, yaml.safe_load(yml_output))

            # Check to make sure we get None with juju status fails
            results = c.status('not-bootstrapped')
            self.assertEqual(results, None)

    @patch('os.path.exists')
    @patch('__builtin__.open')
    def test_conductor_load_envs_yaml(self, mock_open, mock_exists):
        mock_open.return_value.__enter__ = lambda s: s
        mock_open.return_value.__exit__ = Mock()
        mock_open.return_value.read.return_value = RAW_ENVIRONMENTS_YAML
        mock_exists.return_value = True

        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            results = c.load_environments_yaml()
            mock_open.assert_called_once_with(
                os.path.join(
                    os.path.expanduser('~'),
                    '.juju', 'environments.yaml'), 'r')
            self.assertEqual(results, yaml.safe_load(RAW_ENVIRONMENTS_YAML))

    @patch('os.path.exists')
    def test_conductor_load_envs_yaml_error(self, mock_exists):
        mock_exists.return_value = False
        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            self.assertRaises(IOError, c.load_environments_yaml)

    @patch.object(juju_test.Conductor, 'load_environments_yaml')
    def test_get_environment(self, mock_conductor_ley):
        mock_conductor_ley.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)

        juju_env = 'gojuju'
        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            environment = c.get_environment(juju_env)

            self.assertEqual(
                environment,
                PARSED_ENVIRONMENTS_YAML['environments'][juju_env])

    @patch.object(juju_test.Conductor, 'load_environments_yaml')
    def test_get_environment_fails(self, mock_conductor_ley):
        mock_conductor_ley.side_effect = IOError

        juju_env = 'gojuju'
        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)

            self.assertRaises(IOError, c.get_environment, juju_env)

    @patch.object(juju_test.Conductor, 'load_environments_yaml')
    def test_get_environment_notfound(self, mock_conductor_ley):
        mock_conductor_ley.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)

        juju_env = 'dont-exist'
        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)

            self.assertRaises(KeyError, c.get_environment, juju_env)

    @patch('subprocess.check_call')
    def test_conductor_destroy(self, mock_check_call):
        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=1, patch=1)
            good_env = 'valid'
            c.destroy(good_env)

            mock_check_call.assert_called_once_with(
                ['juju', 'destroy-environment',
                 '-y', '-e', good_env], env=c.env
            )

            mock_check_call.reset_mock()
            c.juju_version = juju_test.JujuVersion(major=1, minor=17, patch=0)
            c.destroy(good_env)

            mock_check_call.assert_called_once_with(
                ['juju', 'destroy-environment', '-y', good_env], env=c.env)

            mock_check_call.reset_mock()
            c.juju_version = juju_test.JujuVersion(major=0, minor=8, patch=0)
            c.destroy(good_env)

            expected_cmd = 'echo y | juju destroy-environment -e %s' % good_env
            mock_check_call.assert_called_once_with(
                expected_cmd, shell=True, env=c.env)

    @patch('subprocess.check_call')
    def test_conductor_destroy_failed(self, mock_check_call):
        from subprocess import CalledProcessError
        mock_check_call.side_effect = CalledProcessError('err', 'err', 'err')

        bad_env = 'invalid'

        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=8, patch=0)
            self.assertRaises(juju_test.DestroyUnreliable, c.destroy, bad_env)

            mock_check_call.reset_mock()
            c.juju_version = juju_test.JujuVersion(major=0, minor=8, patch=0)
            self.assertRaises(juju_test.DestroyUnreliable, c.destroy, bad_env)

    @patch('time.sleep')
    @patch('subprocess.check_call')
    @patch.object(juju_test.Conductor, 'status')
    def test_conductor_bootstrap(self, mock_status, mcheck_output, msleep):
        goyml_output = '''
        machines:
            '0':
                agent-state: started
                agent-version: 999.999.999
                dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
                instance-id: j-xxxxxx
                series: world
        services: {}'''

        pyyml_output = '''
        machines:
            0:
                agent-state: running
                instance-state: exists
                dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
                instance-id: j-xxxxxx
                series: world
        services: {}'''

        gostatus_output = yaml.safe_load(goyml_output)
        pystatus_output = yaml.safe_load(pyyml_output)

        mock_status.side_effect = [None, gostatus_output, pystatus_output]
        juju_env = 'test-env'
        expected_cmd = ['juju', 'bootstrap', '-e', juju_env]

        args = Arguments(tests='dummy', upload_tools=False, constraints=False)
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=8, patch=0)

            c.bootstrap(juju_env)
            mcheck_output.assert_called_once_with(expected_cmd, env=c.env)

            # Cover python as well.
            mcheck_output.reset_mock()
            juju_env = 'test-env'
            expected_cmd = ['juju', 'bootstrap', '-e', juju_env]

            args = Arguments(
                tests='dummy', upload_tools=False,
                constraints=False)
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=0, minor=8, patch=0)

            c.bootstrap(juju_env)
            mcheck_output.assert_called_once_with(expected_cmd, env=c.env)

    @patch('time.sleep')
    @patch('subprocess.check_call')
    @patch.object(juju_test.Conductor, 'status')
    def test_conductor_bootstrap_go_opts(self, mstatus, mcheck_output, msleep):
        goyml_output = '''
        machines:
            '0':
                agent-state: started
                agent-version: 999.999.999
                dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
                instance-id: j-xxxxxx
                series: world
        services: {}'''

        gostatus_output = yaml.safe_load(goyml_output)
        mstatus.return_value = gostatus_output
        juju_env = 'test-env'
        expected_cmd = ['juju', 'bootstrap', '--upload-tools', '-e', juju_env]

        args = Arguments(tests='dummy', upload_tools=True, constraints=False)
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=8, patch=0)

            c.bootstrap(juju_env)
            mcheck_output.assert_called_once_with(expected_cmd, env=c.env)

            mcheck_output.reset_mock()
            const_cmd = [
                'juju', 'bootstrap', '--constraints', 'mem=8G,arch=arm',
                '-e', juju_env]

            args = Arguments(tests='dummy', upload_tools=False,
                             constraints='mem=8G,arch=arm')
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=8, patch=0)

            c.bootstrap(juju_env)
            mcheck_output.assert_called_once_with(const_cmd, env=c.env)

    @patch('subprocess.check_call')
    def test_conductor_bootstrap_error(self, mcheck_output):
        from subprocess import CalledProcessError
        mcheck_output.side_effect = CalledProcessError('err', 'err', 'err')

        juju_env = 'bad-env'
        args = Arguments(tests='dummy', upload_tools=True, constraints=False)
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=8, patch=0)

            self.assertRaises(juju_test.BootstrapError, c.bootstrap, juju_env)

    @patch('subprocess.check_call')
    @patch.object(juju_test, 'timeout')
    @patch.object(juju_test.Conductor, 'status')
    def test_conductor_bootstrap_unreliable(self, mock_status, mtimeout,
                                            mcheck_output):
        goyml_output = '''
        machines:
            '0':
                agent-state: pending
                agent-version: 999.999.999
                dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
                instance-id: j-xxxxxx
                series: world
        services: {}'''

        gostatus_output = yaml.safe_load(goyml_output)
        mtimeout.side_effect = juju_test.TimeoutError
        mock_status.return_value = gostatus_output
        jenv = 'test-env'

        args = Arguments(tests='dummy', upload_tools=False, constraints=False)
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=8, patch=0)

            self.assertRaises(
                juju_test.BootstrapUnreliable, c.bootstrap, jenv, 1)

    def test_conductor_isolate_environment(self):
        args = Arguments(tests='dummy')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            self.assertRaises(
                NotImplementedError, c.isolate_environment, 'dummy')

    def test_orchestra_map_status_code(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(tests='dummy')
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')

            self.assertEqual('pass', o.map_status_code(0))
            self.assertEqual('fail', o.map_status_code(1))
            self.assertEqual('skip', o.map_status_code(100))
            self.assertEqual('timeout', o.map_status_code(124))

    def test_orchestra_determine_status_pass(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(tests='dummy', on_timeout='pass')
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')

            self.assertEqual('fail', o.determine_status(1))
            self.assertEqual('skip', o.determine_status(100))
            self.assertEqual('pass', o.determine_status(0))
            self.assertEqual('pass', o.determine_status(124))

    def test_orchestra_determine_status_skip(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(tests='dummy', on_timeout='skip')
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')

            self.assertEqual('pass', o.determine_status(0))
            self.assertEqual('fail', o.determine_status(1))
            self.assertEqual('skip', o.determine_status(100))
            self.assertEqual('skip', o.determine_status(124))

    def test_orchestra_determine_status_fail(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(
                tests='dummy', on_timeout='fail',
                fail_on_skip=True)
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')

            self.assertEqual('pass', o.determine_status(0))
            self.assertEqual('fail', o.determine_status(1))
            self.assertEqual('fail', o.determine_status(100))
            self.assertEqual('fail', o.determine_status(124))

            # Cornercase, timeouts sent to skip
            # should ultimately fail do to args
            o.conductor.args.on_timeout = 'skip'
            self.assertEqual('fail', o.determine_status(124))

    def test_orchestra_print_status_pass(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(tests='dummy', on_timeout='pass')
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')
            o.log.status = MagicMock()

            o.print_status(0)
            o.log.status.assert_called_once_with('%s' % juju_test.TEST_PASS)

            o.log.status.reset_mock()
            o.print_status(124)
            o.log.status.assert_called_with(
                '%s (%s)' % (juju_test.TEST_PASS, juju_test.TEST_TIMEOUT))

    def test_orchestra_print_status_skip(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(tests='dummy', on_timeout='skip')
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')
            o.log.status = MagicMock()

            o.print_status(100)
            o.log.status.assert_called_once_with('%s' % juju_test.TEST_SKIP)

            o.log.status.reset_mock()
            o.print_status(124)
            o.log.status.assert_called_with(
                '%s (%s)' % (juju_test.TEST_SKIP, juju_test.TEST_TIMEOUT))

    def test_orchestra_print_status_fail(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(
                tests='dummy', on_timeout='fail',
                fail_on_skip=True)
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')
            o.log.status = MagicMock()

            o.print_status(1)
            o.log.status.assert_called_once_with('%s' % juju_test.TEST_FAIL)

            o.log.status.reset_mock()
            o.print_status(100)
            o.log.status.assert_called_with('%s (%s)' % (juju_test.TEST_FAIL,
                                                         juju_test.TEST_SKIP))

            o.log.status.reset_mock()
            o.print_status(124)
            o.log.status.assert_called_with(
                '%s (%s)' % (juju_test.TEST_FAIL, juju_test.TEST_TIMEOUT))

            o.conductor.args.on_timeout = 'skip'
            o.log.status.reset_mock()
            o.print_status(124)
            o.log.status.assert_called_with(
                '%s (%s)' % (juju_test.TEST_FAIL, juju_test.TEST_TIMEOUT))

    def test_orchestra_is_passing_code(self):
        with cd('tests_functional/charms/test/'):
            args = Arguments(
                tests='dummy', on_timeout='pass',
                fail_on_skip=False)
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')

            self.assertTrue(o.is_passing_code(0))
            self.assertTrue(o.is_passing_code(100))
            self.assertTrue(o.is_passing_code(124))
            self.assertFalse(o.is_passing_code(1))

            o.conductor.args.on_timeout = 'skip'
            self.assertTrue(o.is_passing_code(124))

            o.conductor.args.on_timeout = 'fail'
            self.assertFalse(o.is_passing_code(124))

            o.conductor.args.on_timeout = 'skip'
            o.conductor.args.fail_on_skip = True
            self.assertFalse(o.is_passing_code(100))
            self.assertFalse(o.is_passing_code(124))

    def test_orchestra_build_env(self):
        with cd('tests_functional/charms/test/'):
            juju_env = 'test'
            args = Arguments(tests='dummy', juju_env=juju_env)
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')
            expected_env = c.env
            expected_env['JUJU_ENV'] = juju_env

            o.build_env()
            self.assertEqual(expected_env, o.env)

    @patch('subprocess.check_call')
    def test_orchestra_rsync_py(self, mcheck_call):
        machine = 0
        juju_env = 'test'
        path = '/var/log/./juju/*'
        logdir = '/tmp/'
        args = Arguments(tests='dummy', juju_env=juju_env)
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=0, minor=8, patch=2)
            o = juju_test.Orchestra(c, 'test/dummy')

            expected_cmd = ['rsync', '-a', '-v', '-z', '-R', '-e',
                            'juju ssh -e %s' % juju_env,
                            '%s:%s' % (machine, path), logdir]
            o.build_env()
            o.rsync(machine, path, logdir)

            mcheck_call.assert_called_once_with(expected_cmd, env=o.env)

    @patch('subprocess.check_call')
    @patch.object(juju_test.Conductor, 'status')
    def test_orchestra_rsync_go(self, mstatus, mcheck_call):
        goyml_output = '''
        machines:
          "0":
            agent-state: started
            agent-version: 1.10.0
            dns-name: juju-yy-yy-yy-yy.compute.cloudprovider.tld
            instance-id: j-yyyyyy
            series: precise
          "1":
            agent-state: started
            agent-version: 1.10.0
            dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
            instance-id: j-xxxxxx
            series: precise
        services:
          dummy:
            charm: cs:precise/dummy-999
            exposed: false
            units:
              dummy/0:
                agent-state: started
                agent-version: 1.10.0
                machine: "1"
                public-address: juju-xx-xx-xx-xx.compute.cloudprovider.tld'''

        goyml_parsed = yaml.safe_load(goyml_output)
        mstatus.return_value = goyml_parsed

        machine = '1'
        dns_name = goyml_parsed['machines']['1']['dns-name']
        juju_env = 'test'
        path = '/var/log/./juju/*'
        logdir = '/tmp/'
        args = Arguments(tests='dummy', juju_env=juju_env)
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=10, patch=0)
            o = juju_test.Orchestra(c, 'test/dummy')

            expected_cmd = ['rsync', '-a', '-v', '-z', '-R', '-e', 'ssh',
                            'ubuntu@%s:%s' % (dns_name, path), logdir]
            o.build_env()
            o.rsync(machine, path, logdir)

            mcheck_call.assert_called_once_with(expected_cmd, env=o.env)

    @patch.object(juju_test.Orchestra, 'rsync')
    @patch.object(juju_test.Conductor, 'status')
    def test_orchestra_archive_logs_go(self, mstatus, rsync):
        goyml_output = '''
        machines:
          "0":
            agent-state: started
            agent-version: 1.10.0
            dns-name: juju-yy-yy-yy-yy.compute.cloudprovider.tld
            instance-id: j-yyyyyy
            series: precise
          "1":
            agent-state: started
            agent-version: 1.10.0
            dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
            instance-id: j-xxxxxx
            series: precise
        services:
          dummy:
            charm: cs:precise/dummy-999
            exposed: false
            units:
              dummy/0:
                agent-state: started
                agent-version: 1.10.0
                machine: "1"
                public-address: juju-xx-xx-xx-xx.compute.cloudprovider.tld'''
        goyml_parsed = yaml.safe_load(goyml_output)
        mstatus.return_value = goyml_parsed

        juju_env = 'testing'
        args = Arguments(tests='dummy', juju_env=juju_env, logdir='/tmp')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=10, patch=0)
            o = juju_test.Orchestra(c, 'test/dummy')

            o.build_env()
            o.archive_logs()
            rsync.assert_has_calls([
                call(
                    0, '/var/./log/juju/*',
                    os.path.join(args.logdir, 'bootstrap', '')),
                call(
                    '1', '/var/./log/juju/*',
                    os.path.join(args.logdir, 'dummy', ''))
            ])

    @patch.object(juju_test.Orchestra, 'rsync')
    @patch.object(juju_test.Conductor, 'status')
    def test_orchestra_archive_logs_py(self, mstatus, rsync):
        pyyml_output = '''
        machines:
          0:
            agent-state: running
            dns-name: juju-yy-yy-yy-yy.compute.cloudprovider.tld
            instance-id: i-yyyyyy
            instance-state: running
          1:
            agent-state: running
            dns-name: juju-xx-xx-xx-xx.compute.cloudprovider.tld
            instance-id: i-xxxxxx
            instance-state: running
        services:
          dummy:
            charm: cs:precise/dummy-19
            relations: {}
            units:
              dummy/0:
                agent-state: started
                machine: 1
                public-address: juju-xx-xx-xx-xx.compute.cloudprovider.tld'''
        pyyml_parsed = yaml.safe_load(pyyml_output)
        mstatus.return_value = pyyml_parsed

        juju_env = 'testing'
        args = Arguments(tests='dummy', juju_env=juju_env, logdir='/tmp')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=0, minor=7, patch=0)
            o = juju_test.Orchestra(c, 'test/dummy')

            o.build_env()
            o.archive_logs()
            rsync.assert_has_calls([
                call(
                    0, '/var/./log/juju/*',
                    os.path.join(args.logdir, 'bootstrap', '')),
                call(
                    1, '/var/./log/juju/*',
                    os.path.join(args.logdir, 'dummy', '')),
                call(
                    1, '/var/lib/juju/units/./*/charm.log',
                    os.path.join(args.logdir, 'dummy', ''))
            ])

    @patch.object(juju_test.Conductor, 'status')
    def test_orchestra_archive_logs_status_fails(self, mstatus):
        mstatus.return_value = None

        args = Arguments(tests='dummy', juju_env='testing', logdir='/tmp')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            o = juju_test.Orchestra(c, 'test/dummy')
            o.build_env()

            self.assertRaises(juju_test.OrchestraError, o.archive_logs)

    @patch.object(juju_test.Orchestra, 'rsync')
    @patch.object(juju_test.Conductor, 'status')
    def test_orchestra_archive_logs_rsync_fails(self, mstatus, rsync):
        from subprocess import CalledProcessError
        mstatus.return_value = {'services': {'dummy': {'units': {'dummy/0':
                                {'machine': '1'}}}}, 'machines': {}}

        args = Arguments(tests='dummy', juju_env='testing', logdir='/tmp')
        with cd('tests_functional/charms/test/'):
            c = juju_test.Conductor(args)
            c.juju_version = juju_test.JujuVersion(major=1, minor=7, patch=0)
            o = juju_test.Orchestra(c, 'test/dummy')
            o.log.warn = MagicMock()
            o.build_env()

            rsync.side_effect = CalledProcessError('err', 'err', 'err')
            o.archive_logs()
            expected_warns = [call('Failed to fetch logs for bootstrap node'),
                              call('Failed to grab logs for dummy/0')]
            o.log.warn.assert_has_calls(expected_warns)


class TestCfgTest(unittest.TestCase):
    test_config = '''\
    options:
      timeout: 500
    substrates:
      order: include, skip
      include: local
      skip: "*"
    '''

    def test_init_str(self):
        t = juju_test.TestCfg(self.test_config)
        self.assertEqual(t.timeout, 500)
        self.assertEqual(t.substrates['include'], 'local')

    def test_init_dict(self):
        t = juju_test.TestCfg(yaml.safe_load(self.test_config))
        self.assertEqual(t.timeout, 500)
        self.assertEqual(t.substrates['include'], 'local')

    def test_update(self):
        t = juju_test.TestCfg(self.test_config)
        t.update(timeout=400)
        self.assertEqual(t.timeout, 400)



#!/usr/bin/python
#
#    create - generate Juju charm from template
#
#    Copyright (C) 2011  Canonical Ltd.
#    Author: Clint Byrum <clint.byrum@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys
import argparse

from charmtools.generators import (
    CharmGenerator,
    CharmGeneratorException,
    get_installed_templates,
)
from . import utils

log = logging.getLogger(__name__)

DEFAULT_TEMPLATE = 'reactive-python'


def setup_parser():
    parser = argparse.ArgumentParser(
        description='create a new charm')

    parser.add_argument(
        'charmname',
        help='Name of charm to create.',
    )
    parser.add_argument(
        'charmhome', nargs='?',
        help='Dir to create charm in. Defaults to CHARM_HOME env var or PWD',
    )
    parser.add_argument(
        '-t', '--template', default=None,
        help='Name of charm template to use; default is ' + DEFAULT_TEMPLATE +
             '. Installed templates: ' + ', '.join(get_installed_templates()),
    )
    parser.add_argument(
        '-a', '--accept-defaults',
        help='Accept all template configuration defaults without prompting.',
        action='store_true', default=False,
    )
    parser.add_argument(
        '-v', '--verbose',
        help='Print debug information',
        action='store_true', default=False,
    )
    utils.add_plugin_description(parser)

    return parser


def main():
    parser = setup_parser()
    args = parser.parse_args()
    args.charmhome = args.charmhome or os.getenv('CHARM_HOME', '.')
    args.config = None

    if args.verbose:
        logging.basicConfig(
            format='%(levelname)s %(filename)s: %(message)s',
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format='%(levelname)s: %(message)s',
            level=logging.INFO,
        )

    if not args.template:
        log.info(
            "Using default charm template (%s). To select a different "
            "template, use the -t option.", DEFAULT_TEMPLATE)
        args.template = DEFAULT_TEMPLATE
    elif args.template not in get_installed_templates():
        raise Exception("No template available for '%s'. Available templates "
                        "may be listed by running 'charm create --help'.")

    generator = CharmGenerator(args)
    try:
        generator.create_charm()
    except CharmGeneratorException as e:
        log.error(e)
        return 1


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python
#
#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import os.path as path
import time
import shutil
import subprocess
import tempfile

from Cheetah.Template import Template
from stat import ST_MODE

from charmtools.generators import (
    CharmTemplate,
)

log = logging.getLogger(__name__)


class AnsibleCharmTemplate(CharmTemplate):
    skip_parsing = ['README.ex', 'Makefile']

    def create_charm(self, config, output_dir):
        self._copy_files(output_dir)

        for root, dirs, files in os.walk(output_dir):
            for outfile in files:
                if outfile in self.skip_parsing:
                    continue

                self._template_file(config, path.join(root, outfile))

        self._cleanup_hooks(config, output_dir)
        self._install_charmhelpers(output_dir)

    def _copy_files(self, output_dir):
        here = path.abspath(path.dirname(__file__))
        template_dir = path.join(here, 'files')
        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        shutil.copytree(template_dir, output_dir)

    def _template_file(self, config, outfile):
        if path.islink(outfile):
            return

        mode = os.stat(outfile)[ST_MODE]
        t = Template(file=outfile, searchList=(config))
        o = tempfile.NamedTemporaryFile(
            dir=path.dirname(outfile), delete=False)
        os.chmod(o.name, mode)
        o.write(str(t))
        o.close()
        backupname = outfile + str(time.time())
        os.rename(outfile, backupname)
        os.rename(o.name, outfile)
        os.unlink(backupname)

    def _cleanup_hooks(self, config, output_dir):
        # Symlinks must be relative so that they are copied properly
        # when output_dir is moved to it's final location.
        for link in ['config-changed', 'install', 'start', 'stop',
                     'upgrade-charm']:
            os.symlink('hooks.py', os.path.join(output_dir, 'hooks', link))

    def _install_charmhelpers(self, output_dir):
        helpers_dest = os.path.join(output_dir, 'lib', 'charmhelpers')
        if not os.path.exists(helpers_dest):
            os.makedirs(helpers_dest)

        cmd = './scripts/charm_helpers_sync.py -c charm-helpers.yaml'
        subprocess.check_call(cmd.split(), cwd=output_dir)
